#!/usr/bin/env python

"""Convert data to an array of colors.

The function `get_color_mapped_values` can translate a series of numeric
values to a Linear or Categorical color map, and can translate an array
including non-numeric data types to a Categorical color map. In all
cases NaN values are mapped to 'transparent'.
"""

from typing import Union, Optional, Sequence, Any, TypeVar
from loguru import logger
import numpy as np
import toyplot
from toytree.color import ToyColor
from toytree.utils.src.exceptions import ToytreeError

logger = logger.bind(name="toytree")
ToyTree = TypeVar("ToyTree")


CMAP_ERROR = """
To find a list of valid colormap names supported by toyplot use:
>>> print(toyplot.color.brewer.names())

You can map feature values to colors by entering (feature, colormap):
>>> tree.draw('c', node_colors=('dist', 'Spectral'))

or, load a colormap as a Map object for finer control:
>>> cmap = toyplot.color.brewer.map('BlueRed', domain_min=0, domain_max=1)
>>> tree.draw('c', node_colors=('dist', cmap))

(See https://toyplot.readthedocs.io/en/stable/colors.html#Color-Maps)\
"""

CATEGORICAL_TOO_FEW_CATEGORIES = """Error in (feature, colormap)
Data contains more unique values than contained in the selected CategoricalMap
colormap. Consider selecting a LinearMap instead, such as 'Spectral'.

(See https://toyplot.readthedocs.io/en/stable/colors.html#Color-Maps)\
"""

PALETTE_TOO_FEW_CATEGORIES = """Error in (feature, colormap)
Data contains more unique values than contained in the selected Palette.
Consider selecting a LinearMap instead, such as 'Spectral'.

(See https://toyplot.readthedocs.io/en/stable/colors.html#Color-Maps)\
"""


def get_color_mapped_feature(
    tree: ToyTree,
    feature: str,
    cmap: Union[str, toyplot.color.Map] = None,
    domain_min: Optional[float] = None,
    domain_max: Optional[float] = None,
    nan_value: Optional[float] = None,
    tips_only: bool = False,
    reverse: bool = False,
) -> np.ndarray:
    """Return an array of colors mapped to feature data in a tree.

    Parameters
    ----------
    tree: ToyTree
        A ToyTree to extract feature data from.
    feature: str
        Name of a feature to extract from one or more Nodes in the tree.
    cmap: str or toyplot.color.Map
        A toyplot colormap object or a str name of a colormap.
    domain_min: float or None
        If None the domain will be automatically fit to the data min.
    domain_max: float or None
        If None the domain will be automatically fit to the data max.
    nan_value: float or None
        A value to substitute for any missing (NaN) values. If None
        then NaN values will be mapped to 'transparent'.
    tips_only: bool
        If True then data is only projected and returned for tip Nodes.
    reverse: bool
        Reverse the order of the colormap.
    """
    assert hasattr(tree, "nnodes"), "first argument must be a ToyTree."
    if tips_only:
        values = tree.get_tip_data(feature).values
    else:
        values = tree.get_node_data(feature).values
    return get_color_mapped_values(values, cmap, domain_min, domain_max, nan_value, reverse)


def get_color_mapped_values(
    values: Sequence[Any],
    cmap: Union[str, toyplot.color.Map] = None,
    domain_min: Optional[float] = None,
    domain_max: Optional[float] = None,
    nan_value: Optional[float] = None,
    reverse: bool = False,
) -> np.ndarray:
    """Return feature mapped to a continuous or discrete color map.

    Raises helpful error messages if user entered the special tuple
    argument incorrectly, or by accident.

    Parameters
    ----------
    values: Series[float]
        A Series of numeric values.
    cmap: str or toyplot.color.Map
        A toyplot colormap object or a str name of a colormap.
    domain_min: float or None
        If None the domain will be automatically fit to the data min.
    domain_max: float or None
        If None the domain will be automatically fit to the data max.
    nan_value: float or None
        A value to substitute for any missing (NaN) values. If None
        then NaN values will be mapped to 'transparent'.
    reverse: bool
        Reverse the order of the colormap.

    Examples
    --------
    >>> ...
    """
    # Use Spectral as default map if None provided.
    if cmap is None:
        cmap = "Spectral"

    # store original cmap var before conversion to ColorMap
    cmap_orig = cmap

    # if colormap is str then expand to a color.Map
    if isinstance(cmap, str):
        # return as Categorial or LinearMap or...
        try:
            cmap = toyplot.color.brewer.map(cmap, domain_min=domain_min, domain_max=domain_max, reverse=reverse)
        except KeyError:
            # return as LinearMap
            try:
                cmap = toyplot.color.linear.map(cmap, domain_min=domain_min, domain_max=domain_max)
            except KeyError:
                # return as DivergingMap
                try:
                    cmap = toyplot.color.diverging.map(cmap, domain_min=domain_min, domain_max=domain_max, reverse=reverse)
                except KeyError:
                    msg = "Invalid colormap arg for (feature, colormap) input.\n" + CMAP_ERROR
                    raise ToytreeError(msg)

    # if colormap is not Map or Palette (good) then check for messier formats
    if not isinstance(cmap, (toyplot.color.Map, toyplot.color.Palette)):

        # try converting colormap to a Palette for ndarray or Sequence[Color]
        # and add a transparent color at end to use for np.nan.
        try:
            cmap = toyplot.color.Palette(cmap, reverse=reverse)
            cmap += toyplot.color.Palette(["transparent"])
        except Exception:
            msg = "Invalid colormap arg for (feature, colormap) input.\n" + CMAP_ERROR
            raise ToytreeError(msg)

    # --- colormap is now CategoricalMap, LinearMap or Palette ---

    # if map is Categorical then try to treat data as categorical, or raise error
    if isinstance(cmap, toyplot.color.CategoricalMap):
        # convert data to str, with nan as highest sort str character
        cvalues = [str(i) for i in values]
        cvalues = ["~~" + i if i == "nan" else i for i in cvalues]

        # convert str data to int categories
        categories, cvalues = np.unique(cvalues, return_inverse=True)

        # raise warning if not enough categories for data
        if cmap.domain.max < len(categories):
            # logger.warning(CATEGORICAL_TOO_FEW_CATEGORIES)
            raise ToytreeError(CATEGORICAL_TOO_FEW_CATEGORIES)

        # recreate the map using count of current data
        if isinstance(cmap_orig, str):
            # handle toyplot error that does not allow ncategories=2
            try:
                cmap = toyplot.color.brewer.map(cmap_orig, count=len(categories), domain_min=domain_min, domain_max=domain_max, reverse=reverse)
            # ...HERE...
            except KeyError:
                cmap = toyplot.color.brewer.map(cmap_orig, count=max(3, len(categories)), domain_min=domain_min, domain_max=domain_max, reverse=reverse)                

        # broadcast data to color map
        colors = cmap.colors(cvalues)

    # if cmap is a Palette
    elif isinstance(cmap, toyplot.color.Palette):

        # convert data to int categories
        cvalues = [str(i) for i in values]
        categories, cvalues = np.unique(cvalues, return_inverse=True)

        # raise warning if not enough categories for data
        if len(cmap) < len(categories):
            raise ToytreeError(PALETTE_TOO_FEW_CATEGORIES)

        # broadcast data to color map
        colors = np.array([cmap.color(i) for i in cvalues])

    # if map is Linear (data can be linear or categorial)
    else:

        # convert values to floats and fit colormap to data domain
        try:
            cvalues = np.array(values, dtype=float)

        # but if data is categorical then convert data to ints
        except (TypeError, ValueError):
            cvalues = [str(i) for i in values]
            categories, cvalues = np.unique(cvalues, return_inverse=True)

            # recreate the map using count of current data
            try:
                if isinstance(cmap_orig, str):
                    cmap = toyplot.color.brewer.map(cmap_orig, count=len(categories), domain_min=domain_min, domain_max=domain_max, reverse=reverse)
            # try to convert linearmap to categorical
            except KeyError:
                cmap = toyplot.color.brewer.map(cmap_orig, count=max(3, len(categories) + 1), domain_min=domain_min, domain_max=domain_max, reverse=reverse)                
                # raise ToytreeError(
                #     "Data cannot be colormapped. You selected discrete data but a linear colormap.\n"
                #     f"  data={cvalues[:5]}...\n"
                #     f"  cmap={cmap}.\n"
                #     "Try an alternative cmap such as 'Set2'.\n\n"
                #     "(See https://toyplot.readthedocs.io/en/stable/colors.html#Color-Maps)"
                # ) from exc

        # fit colormap domain to float values or int categories
        if not all(np.isnan(cvalues)):
            if cmap.domain.min is None:
                cmap.domain.min = np.nanmin(cvalues)
            if cmap.domain.max is None:
                cmap.domain.max = np.nanmax(cvalues)

        # broadcast values to color map
        colors = cmap.colors(cvalues)

    # replace color w/ "transparent" for any nan data points. BTW, Broadcasting
    # can't be used here b/c we converted values to a list earlier.
    for idx, value in enumerate(values):
        try:
            if np.isnan(value):
                if nan_value is None:
                    colors[idx] = ToyColor((0, 0, 0, 0))
                else:
                    colors[idx] = cmap.color(nan_value)
                # logger.warning(f"NAN color: {colors[idx]}")
        # skip if type=str or other that cannot be checked for nan.
        except TypeError:
            pass
    return colors


if __name__ == "__main__":

    DATA = [
        np.linspace(0, 1, 7),
        [1, 1, 1, 0, 0, 0, 0],
        [1] * 3 + [np.nan] * 4,
        list("aaabbbb"),
        ['a'] * 3 + [0] + [np.nan] * 3,
    ]

    COLOR_MAPS = [
        "BlueRed",
        "Spectral",
        toyplot.color.brewer.map("BlueRed"),
        "Set2",
    ]

    for idx, values in enumerate(DATA):
        for cmap in COLOR_MAPS:
            print(f"dataset={idx}, colormap={cmap}")
            print(get_color_mapped_values(values, cmap))



#!/usr/bin/env python

"""Inference methods concerning multi-labeled trees (mul trees).

Count the number of duplications and losses required to embed a gene
tree into a species tree.

References
----------
...
"""

from typing import Dict, Optional, Sequence
import pandas as pd
import toytree
from toytree import ToyTree

# ToyTree = TypeVar("ToyTree")
# Node = TypeVar("Node")


def _set_ng_labels(gtree: ToyTree) -> None:
    """Sets a cached feature 'ng' with leaf names under each node.

    In the example below there are two gene tree tips named 'r1'.

    Example
    -------
    >>> tree = toytree.rtree.unittree(ntips=5, seed=123)
    >>> tree1 = tree.mod.add_internal_node_and_child("r1", name="r1")
    >>> _set_ng_labels(tree1)
    >>> tree1.get_node_data("ng")
    >>> # 0                     {r0}
    >>> # 1                     {r1}
    >>> # 2                     {r1}
    >>> # 3                     {r2}
    >>> # 4                     {r3}
    >>> # 5                     {r4}
    >>> # 6                     {r1}
    >>> # 7                 {r1, r0}
    >>> # 8             {r1, r0, r2}
    >>> # 9                 {r3, r4}
    >>> # 10    {r4, r3, r2, r1, r0}
    """
    for node in gtree:
        if node._children:
            node.ng = set.union(*(i.ng for i in node._children))
        else:
            node.ng = {node.name}


def _set_ns_labels(gtree: ToyTree, sptree: ToyTree) -> None:
    """Sets feature 'ns' as MRCA sptree node idx containing 'ng' set.

    This assumes that `_set_ng` has already set 'ng' features on the
    gene tree. The stored 'ns' features are species tree idxs.

    Example
    -------
    >>> tree = toytree.rtree.unittree(ntips=5, seed=123)
    >>> tree1 = tree.mod.add_internal_node_and_child("r1", name="r1")
    >>> _set_ng_labels(tree1)
    >>> _set_ns_labels(tree1, tree)
    >>> tree1.get_node_data("ns")
    >>> # 0     0
    >>> # 1     1
    >>> # 2     1
    >>> # 3     2
    >>> # 4     3
    >>> # 5     4
    >>> # 6     1
    >>> # 7     5
    >>> # 8     6
    >>> # 9     7
    >>> # 10    8
    """
    for node in gtree:
        node.ns = sptree.get_mrca_node(*node.ng).idx


def _count_duplications(gtree: ToyTree) -> int:
    """Return numer of duplication events on a gene tree.

    Nodes in the gene tree are said to be duplication nodes when they
    map to the same species tree node as at least one of their
    descendants. This assumes that the gene trees already have a
    value "ns" stored to every Node by `_set_ns_labels`.

    Adds a feature 'dup' to Nodes with True or False.

    Note
    ----
    - Can you have multiple duplications per node?
    - How to treat polytomies?

    Example
    -------
    >>> tree = toytree.rtree.unittree(ntips=5, seed=123)
    >>> tree1 = tree.mod.add_internal_node_and_child("r1", name="r1")
    >>> _set_ng_labels(tree1)
    >>> _set_ns_labels(tree1, tree)
    >>> _count_duplications(tree1)
    >>> # 1
    >>> tree1.get_node_data("dup")
    >>> # 0     False
    >>> # 1     False
    >>> # 2     False
    >>> # 3     False
    >>> # 4     False
    >>> # 5     False
    >>> # 6      True
    >>> # 7     False
    >>> # 8     False
    >>> # 9     False
    >>> # 10    False
    """
    ndups = 0
    for node in gtree[gtree.ntips:]:
        # if any descendants (not including self) trace back to the
        # same species tree interval then a duplication occurred
        node.duplication = False
        for desc in node.iter_descendants():
            if node == desc:
                continue
            if node.ns == desc.ns:
                ndups += 1
                node.duplication = True
                break
    return ndups


def _count_losses(gtree: ToyTree, sptree: ToyTree) -> int:
    """Return number of loss events on a gene tree.

    Nodes in the gene tree count a loss for every occurrence
    descendants do not include are descended
    from a duplication event but do not include all descendants
    leaves.

    Adds feature 'loss' as an int to each Node.
    """
    losses = 0

    # measure number of nodes between the sptree root and the min
    # sptree interval in which a gtree node must have occurred.
    for node in gtree:
        node.depth = sptree.distance.get_node_distance(
            -1, node.ns, topology_only=True) + 1

    # measure losses as ...
    for node in gtree[:-1]:
        node.losses = ((node.depth - node.up.depth) - 1) + int(node.up.duplication)
        if node.losses:
            losses += 1
    return losses


def get_duplication_loss_coalescence(
    gtree: ToyTree, sptree: ToyTree, imap: Optional[Dict[str, Sequence[str]]] = None,
) -> Dict[str, int]:
    """Returns a dict with duplication, loss, and coalescence info.

    ...

    Parameters
    ----------
    gtree: ToyTree
        ...
    sptree: ToyTree
        ...
    imap: Dict[str, Sequence[str]]
        A dict mapping species tree tip labels to a list of one or
        more gene tree tip labels. If not provided (None) it is
        expected that the names will match identically.

    Example
    -------
    >>> ...
    >>> ...
    """
    if imap:
        raise NotImplementedError("TODO")
    gtree = gtree.copy()  # speed cost but prevents added features.
    _set_ng_labels(gtree)
    _set_ns_labels(gtree, sptree)
    duplications = _count_duplications(gtree)
    losses = _count_losses(gtree, sptree)
    data = {
        "tree": gtree,
        "duplications": duplications,
        "losses": losses,
        # "coalescences": gtree.nnodes - dups - loss,
        "score": duplications + losses,
    }
    return data


def get_multree_reconciliation_scores(
    gtrees: ToyTree | Sequence[ToyTree],
    sptrees: ToyTree | Sequence[ToyTree],
) -> pd.DataFrame:
    """Return table with DLC reconciliation...

    Parameters
    ----------
    ...

    Returns
    -------
    A pandas dataframe is returned w/ ...[sptree, gtree, dtg, ltg, score]

    Example
    --------
    >>> tree = toytree.rtree.unittree(ntips=5, seed=123)
    >>> dtree = tree.mod.add_internal_node_and_child('r3', name='r3')
    >>> get_multree_reconciliation_score(tree, dtree)
    """
    # check that gtrees and sptrees are iterable (lists)
    if isinstance(gtrees, ToyTree):
        gtrees = [gtrees]
    if isinstance(sptrees, ToyTree):
        sptrees = [sptrees]

    # set 'ng' sets on the Nodes of all gene trees.
    for gtree in gtrees:
        _set_ng_labels(gtree)

    # build collection of species trees to be tested...
    # ...

    # iterate over species trees to compute scores. Linear in time,
    # can be fully parallelized, but is not yet.
    full_data = []
    sub_data = []
    for sidx, sptree in enumerate(sptrees):
        newick_s = sptree.write(None, None, None)

        # cache the MRCA nodes
        for gtree in gtrees:
            _set_ns_labels(gtree, sptree)

        # iterate over ...
        chunks = []
        for gidx, gtree in enumerate(gtrees):
            newick_g = gtree.write(None, None, None)
            dtg = _count_duplications(gtree)
            ltg = _count_losses(gtree, sptree)
            score = dtg + ltg
            chunks.append([sidx, gidx, newick_s, newick_g, dtg, ltg, score])

        # fill result of all gene trees
        sub_data.append(
            pd.DataFrame(
                chunks,
                columns=['sidx', 'gidx', 'sptree', 'gtree', 'dups', 'losses', 'score'],
            ))
        full_data.append([sidx, newick_s, sub_data[-1].score.sum()])

    # concatenate dataframes into final df and relabel
    sub_data = pd.concat(sub_data).reset_index(drop=True)
    full_data = pd.DataFrame(full_data, columns=["sidx", "sptree", "score"])
    full_data = full_data.sort_values("sidx")
    return full_data, sub_data


if __name__ == "__main__":


    tree = toytree.rtree.unittree(5, seed=123)
    tree1 = tree.mod.add_internal_node_and_child("r1", name="r1")
    _set_ng_labels(tree1)
    print(tree1.get_node_data("ng"))

    _set_ns_labels(tree1, tree)
    print(tree1.get_node_data("ns"))
    # validation with ipcoal
    # import ipcoal
    # import string

    # # get a single-labeled species tree with names (A-...)
    # NTIPS = 6
    # sptree_single = toytree.rtree.unittree(ntips=NTIPS, treeheight=1e6, seed=123)
    # name_map = dict(zip(range(10), string.ascii_uppercase))
    # sptree_single.set_node_data("name", name_map, inplace=True)

    # # insert an allo-polyploid genome dup event to make a mul-tree
    # subtree = toytree.tree("((X:1,Y:1):1,Z:2):1;")
    # sptree_mul = toytree.mod.add_internal_node_and_subtree(sptree_single, 'A', subtree=subtree)
    # sptree_mul = toytree.mod.add_internal_node_and_subtree(sptree_mul, 'E', subtree=subtree)

    # # generate a different mul-tree to compare against
    # sptree_mul2 = toytree.mod.add_internal_node_and_subtree(sptree_single, 'B', subtree=subtree)
    # sptree_mul2 = toytree.mod.add_internal_node_and_subtree(sptree_mul2, 'F', subtree=subtree)

    # # draw species trees
    # c, _, _ = toytree.mtree([sptree_mul, sptree_mul2]).draw(ts='s', height=400)
    # toytree.utils.show(c, tmpdir="~")

    # # sptree_mul2

    # simulate genealogies on this mul-species-tree
    # model = ipcoal.Model(
    #     sptree_mul, Ne=2e5, nsamples=1, seed_trees=123, seed_mutations=123)
    # model.sim_trees(100)

    # # draw some genealogies
    # # canvas, _, _ = model.draw_genealogies(scale_bar=True, shared_axes=True)
    # # toytree.utils.show(canvas)
    # gtrees = toytree.mtree(model.df.genealogy.to_list())
    # full, sub = get_multree_reconciliation_score(gtrees, [sptree_mul, sptree_mul2])

    # print("Comparing species tree models:")
    # print("------------------------------")
    # print(full)

    # print("\n\nScores of individual gene trees for each model:")
    # print("--------------------------------------------------")
    # print(sub)



#!/usr/bin/env python

"""(Experimental) Network class. NOT YET IMPLEMENTED.

- network parsing functions
    - bpp format
    - snaq format

- constrained re-rooting of Network

- conversions to and from tree + admixture args.

- drawing function.
"""

from typing import Union, Tuple, List, Optional
from pathlib import Path
import re
from loguru import logger
import toytree
from toytree.io.src.parse import WHITE_SPACE
from toytree.core.node import Node

logger = logger.bind(name="toytree")


class Node3(Node):
    """Nodes contained within a Network class object.

    In contrast to treenodes these nodes can have degree > 3, by
    having >1 parent.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._up: Tuple[Node] = tuple()
        """: Parent is a tuple of 0 or more Node3s."""
        self._gamma: Tuple[float] = tuple()
        """: Ancestry this Node3 received from parent Node3s, in order."""

    # overwrite some functions such as delete, etc.


class ToyNet(toytree.ToyTree):
    """Child class of ToyTree that stores and draws network edges.

    A network object can be used to draw networks, to calculate or
    report statistics on a network, and to extract minor or major
    trees from a network (as ToyTree objects).

    Algorithm to traverse nodes of a network in order... could still
    use the same traversals that trace the major or minor tree to visit
    each node, but another traversal may be available that would also
    visit nodes connected by an admix edge sooner...
    """
    def __init__(self, network: Union[str, Path]):
        self.network = network
        # tree, admix = parse_network(net, disconnect=disconnect)
        # super().__init__(tree, *args, **kwargs)
        # self.admix = admix

    # def ... , disconnect: bool = True, **kwargs):

    def add_edge(self) -> toytree.ToyTree:
        """Return modified Network with a new degree-3 edge added."""

    def remove_edge(self) -> toytree.ToyTree:
        """Return modified Network with a degree-3 edge removed."""

    def get_major_tree(self) -> toytree.ToyTree:
        """Return the major tree..."""

    def get_admixture_edges(self) -> Tuple:
        """Return Tuple with list of admixture edges..."""

    def write(self):
        """Write network in extended format..."""

    def root(self):
        """raise rootMisMatchError if direction doesn't work..."""





#!/usr/bin/env python

"""MultiTree classes for representing a collection of trees.

A MultiTree class is used to store a collection of trees, and includes
functions for visualizing and manipulating multiple trees.

Methods
-------
- get_consensus_tree
    Return a consensus tree from a set of input trees.
- draw
    Return a toyplot drawing of multiple trees spaced on a grid.
- draw_cloud_tree
    Return a toyplot drawing of multiple trees overlaid.

See Also
--------
- toytree.utils.ToyTreeSequence
    A wrapper around the tskit TreeSequence class. This class has
    convenience functions for extracting and plotting trees from a
    TreeSequence object. To use it requires that you install the
    additional library tskit. By contrast the MultiTree class is a
    much simpler collection of trees, and does not require any third
    party packages, and is thus the default object in toytree for
    working with multiple trees.

TODO:
    - set-like function that drops tips from each tree not shared by all trees.
    - root function that drops tree that cannot be rooted or not set-like
    - consensus function drop tips that are not in set.
    - A distinct multitree mark for combining styles on topologies
    - New mark: leave proper space for tips to mtree fits same as tree.
"""

from __future__ import annotations
from typing import Union, List, Sequence, Optional, Tuple, TypeVar, Iterator
from copy import deepcopy
import numpy as np
# import pandas as pd
from loguru import logger
from toytree.utils import ToytreeError
from toytree.core import ToyTree, Node
from toytree.style import TreeStyle, get_base_tree_style_by_name, tree_style_to_css_dict
from toytree.drawing.src.setup_canvas import get_canvas_and_axes
from toytree.drawing.src.setup_grid import Grid
from toytree.drawing.src.draw_cloudtree import draw_cloudtree
from toytree.drawing.src.draw_multitree import draw_multitree
from toytree.annotate import add_axes_scale_bar
# from toytree.core.drawing.render import ToytreeMark
from toytree.infer.src.consensus import ConsensusTree

# from toytree.utils import ToytreeError

# aliases of Toyplot types returned by draw functions
Canvas = TypeVar("Canvas")
Cartesian = TypeVar("Cartesian")
Mark = TypeVar("Mark")
Query = TypeVar("Query", str, int, Node)
logger = logger.bind(name="toytree")


class MultiTree:
    """MultiTree class to visualize and analyze collections of trees.

    MultiTree objects can be indexed to extract ToyTrees, and are also
    iterable. They include functions for visualizing sets of trees, and
    for comparing and calculating statistics on sets of trees.

    Notes
    -----
    Use factory function `toytree.mtree` to init a MultiTree instance
    from a list of ToyTrees, or from a newick or nexus trees file.

    Examples
    --------
    >>> trees = [toytree.rtree.unittree(10) for i in range(100)]
    >>> mtree = toytree.mtree(trees)
    >>> mtree.draw();
    """
    def __init__(self, treelist: List[ToyTree]):
        self.treelist: List[ToyTree] = treelist
        """List of ToyTree objects in the MultiTree."""

        # self.data: pd.DataFrame = self._init_data(treelist, data)
        # """DataFrame with tree metadata (e.g., ipcoal.Model.df)."""

    # def _init_data(self, treelist, data: Optional[pd.DataFrame]):
    #     if data is None:
    #         data = pd.DataFrame(
    #             index=range(len(self)),
    #             data={"count": 1},
    #         )
    #     return data

    def __len__(self):
        """Return number of trees in the treelist"""
        return len(self.treelist)

    def __iter__(self) -> Iterator[ToyTree]:
        """Generator to yield ToyTrees from MultiTree.treelist."""
        for tree in self.treelist:
            yield tree

    def __getitem__(self, idx: int) -> ToyTree:
        """Return ToyTree by indexing from MultiTree.treelist."""
        return self.treelist[idx]

    def __repr__(self):
        """string representation shows ntrees and type"""
        return f"<toytree.MultiTree ntrees={len(self)}>"

    @property
    def ntips(self):
        """Return number of tips in the FIRST tree in treelist."""
        return self.treelist[0].ntips

    @property
    def ntrees(self):
        """Return number of trees in the MultiTree.treelist"""
        return len(self.treelist)

    def all_tree_tip_labels_same(self) -> bool:
        """Return True if names are the same in all the trees."""
        first = set(self.treelist[0].get_tip_labels())
        return all(set(i.get_tip_labels()) == first for i in self)

    def all_tree_topologies_same(self, include_root: bool = False) -> bool:
        """Return True if all topologies in treelist are identical."""
        return len(set(i.get_topology_id(include_root=include_root) for i in self)) == 1

    def all_tree_tips_aligned(self, rtol: float = 1e-5, atol: float = 1e-5) -> bool:
        """Return True if all tree tips are aligned (i.e., ultrametric)

        This checks that all tip Node heights align at 0 within the
        accepted tolerance level.

        Parameters
        ----------
        rtol: float
            See np.allclose() function docstring.
        atol: float
            See np.allclose() function docstring.
        """
        return np.allclose(
            a=np.concatenate([[i.height for i in tree] for tree in self]),
            b=0,
            rtol=rtol,
            atol=atol,
        )

    # def iter_unique_topologies(self, include_root: bool = False) -> Iterator[ToyTree]:
    #     """Generator to yield unique tree topologies in treelist."""
    #     counter = Counter()
    #     for tree in self:
    #         hashed = tree.get_topology_id(include_root=include_root)
    #         counter[hashed] += 1
    #     return counter

    # better name?
    def get_unique_topologies(self, include_root: bool = False) -> List[Tuple[ToyTree, int]]:
        """Return a list of (ToyTree, count) for each unique tree.

        This can be useful for calculating statistics only on the
        unique set of trees and multiplying by their frequency (for
        example this is done when generating consensus trees).

        Parameters
        ----------
        include_root:
            If False then all unique rooted trees are counted and
            returned rather than all unique unrooted trees.

        Examples
        --------
        >>> mtree = toytree.mtree(
        >>>     [toytree.rtree.rtree(6) for i in range(100)])
        >>> print(mtree.get_unique_topology_counts())
        >>> # [(ToyTree, 10), (ToyTree, 9), (ToyTree, 9), ...]
        """
        trees_dict = {}
        for tree in self:
            hashed = tree.get_topology_id(include_root=include_root)
            if hashed in trees_dict:
                trees_dict[hashed][1] += 1
            else:
                trees_dict[hashed] = [tree, 1]
        # sort trees and return as a List of Tuples
        return sorted(trees_dict.values(), key=lambda x: x[1], reverse=True)

    def copy(self):
        """Return a deepcopy of the MultiTree."""
        return deepcopy(self)

    # todo: use wrap
    def write(
        self,
        path: Optional[str] = None,
        dist_formatter: str = "%.12g",
        internal_labels: Optional[str] = "support",
        internal_labels_formatter: Optional[str] = "%.12g",
        features: Optional[Sequence[str]] = None,
        features_prefix: str = "&",
        features_delim: str = ",",
        features_assignment: str = "=",
        **kwargs,
    ) -> Optional[str]:
        """Write tree to newick string and return or write to filepath.

        The newick string can be formatted in several ways. The default
        will include dist values (edge lengths) and support values as
        internal node labels. The edge lengths can be suppressed by
        setting `dist_formatter=None`, and internal node labels can be
        similarly suppressed, or set to store a different feature, such
        as internal node names. Additional features can be stored in the
        node comment blocks in extended-newick-format (NHX-like) by using
        the "features" arguments (see examples).

        Parameters
        ----------
        tree: ToyTree
            A ToyTree instance to write as a newick string.
        path: str or None
            A filepath to write to file, or None to return newick string.
        dist_formatter: str or None
            A formatting string to format float dist values (edge lengths),
            or None to not write dist values. Default is "%.6g".
        internal_labels: str or None
            A feature to write as internal node labels. None suppresses
            internal labels. The 'support' feature is default, and
            often used here, but 'name' or any other feature can be
            used as well.
        internal_labels_formatter: str or None
            A formatting string to format internal labels. If an internal
            label cannot be formatted due to TypeError (e.g., you select
            'name' for `internal_labels` but leave this optional at its
            default as a float formatter '%.6g', instead of str formatter)
            it will simply be converted to a string.
        features: List[str]
            A list of additional features to write in the newick comment
            block. For example, features=["height"] will save heights.
        features_prefix: str
            A prefix character written to the start of newick comment
            blocks. Typical values are "&" (default) or "&&NHX:".
        features_delim: str
            A character used to delimit features in the newick comment
            block. Default is ",".
        features_assignment: str
            A character used to separate feature keys and values. Default
            is "=".

        See Also
        --------
        `write_nexus`
            Write tree newick string in a NEXUS format.
        `ToyTree.write`
            This function is available from ToyTree objects as `.write`.
        """
        if kwargs:
            logger.warning(
                f"Deprecated args to write(): {list(kwargs.values())}. See docs.")
        newicks = []
        for tree in self:
            newicks.append(tree.write(
                path=None, dist_formatter=dist_formatter,
                internal_labels=internal_labels,
                internal_labels_formatter=internal_labels_formatter,
                features=features,
                features_prefix=features_prefix,
                features_delim=features_delim,
                features_assignment=features_assignment,
            ))
        newicks = "\n".join(newicks)
        if path is None:
            return newicks
        with open(path, 'w', encoding="utf-8") as out:
            out.write(newicks)
        return None

    def get_consensus_tree(
        self,
        best_tree: ToyTree = None,
        majority_rule_min: float = 0.0,
        ultrametric: Optional[bool] = None,
    ) -> ToyTree:
        """Return an extended majority rule consensus Toytree.

        Consensus tree Node 'support' features record the frequency of
        occurrence of clades across the input MultiTree treelist.
        Clades with support below 'majority_rule_min' are collapsed
        into polytomies. If you enter an optional 'best_tree' then
        support values from the input trees will be calculated for
        clades in this tree, and the 'best_tree' is returned with
        support values added to Nodes, else a majority-rule consensus
        tree is generated and returned. The mean, min, and max of
        additional features in the trees can also be calculated.

        Parameters
        ----------
        best_tree: Toytree, str, or None
            A tree that support values should be calculated for and
            added to. For example, you want to calculate how often
            clades in your best ML tree are supported in 100 bootstrap
            trees.
        majority_rule_min: float
            Cut-off below which clades are collapsed in the majority
            rule consensus tree. This is a proportion (e.g., 0.5 means
            50%).

        Examples
        --------
        >>> best = toytree.rtree.unittree(ntips=10, seed=123)
        >>> trees = [toytree.rtree.unittree(ntips=10) for i in range(10)]
        >>> mtree = toytree.mtree(trees)
        >>> ctree1 = mtree.get_consensus_tree(trees, best_tree=best)
        >>> ctree2 = mtree.get_consensus_tree(trees)
        >>> ctree3 = mtree.get_consensus_tree(trees, majority_rule_min=0.5)
        >>> toytree.mtree([ctree1, ctree2, ctree3]).draw();
        """
        cons = ConsensusTree(
            mtree=self,
            best_tree=best_tree,
            majority_rule_min=majority_rule_min,
            ultrametric=ultrametric,
        )
        return cons.run()

    ################################################################
    # Tree Modification functions
    # These visit and possibly modify every tree in treelist
    ################################################################

    def root(
        self,
        *query: Query,
        root_dist: Optional[float] = None,
        edge_features: Optional[Sequence[str]] = None,
        inplace: bool = False,
    ) -> MultiTree:
        """Return a MultiTree with all ToyTrees in treelist rooted.

        If a tree cannot be rooted on the selected position this
        will raise a ToytreeError.

        Parameters
        ----------
        *query: str, int, or Node
            One or more Node selectors, which can be Node objects, names,
            or int idx labels. If multiple are entered the MRCA node will
            be used as the base of the edge to split.
        regex: bool
            If True then Node name strings are treated as regular
            expressions that can match to multiple Nodes.
        root_dist: None or float
            The length (dist) along the root edge above the Node query
            where the new root edge should be placed. Default is None
            which will place root at the midpoint of the edge. A float
            can be entered, but will raise ToyTreeError if > len of edge.
        edge_features: Sequence[str]
            One or more Node features that should be treated as a feature
            of its edge, not the Node itself. On rooting, edge features
            are re-polarized, to apply to the correct Node. The 'dist'
            and 'support' features are always treated as edge features.
            Add additional edge features here. See docs for example.
        inplace: bool
            If True the original tree is modified and returned, otherwise
            a modified copy is returned.
        """
        mtree = self if inplace else self.copy()
        for tree in mtree:
            tree.root(
                *query,
                root_dist=root_dist,
                edge_features=edge_features,
                inplace=True,
            )
        return mtree

    def unroot(self, inplace: bool = False) -> MultiTree:
        """Return a MultiTree with all ToyTrees in treelist unrooted"""
        mtree = self if inplace else self.copy()
        for tree in mtree:
            tree.unroot(inplace=True)
        return mtree

    ################################################################
    # Tree Comparison/Distance functions
    # >>> toytree.distance.get_treedist_rf(mtre[0], mtre[1])
    # >>> toytree.distance.get_treedist_matrix(*mtre.treelist)
    ################################################################

    # def get_tree_distance(self, idx0: int, idx1: int, metric: str = "rf") -> float:
    #     """Return a distance metric comparing two trees.

    #     Trees are indexed from the treelist by indices idx0 and idx1.

    #     Parameters
    #     ----------
    #     idx0: int
    #         Index of a tree from the treelist.
    #     idx1: int
    #         Index of a tree from the treelist.
    #     metric: str
    #         Name of a supported tree distance metric. For available
    #         options see `toytree.distance.treedist`.
    #     **kwargs: Dict
    #         Additional options accepted by tree distance method.

    #     Examples
    #     --------
    #     >>> ...
    #     """
    #     return get_tree_distance(
    #         self[idx0], self[idx1], metric=metric, **kwargs)

    # def get_tree_distance_matrix(self, ) -> float:
    #     """TODO..."""

    # def get_tree_distance_distribution(
    #     self, ) -> np.ndarray:
    #     """Return a distribution of tree distances.

    #     Tree distances are measure between pairs of trees, but several
    #     options are available for how pairs will be sampled, including
    #     'random', 'consensus', or 'distance'.

    #     Parameters
    #     ----------

    #     Returns
    #     -------
    #     pd.DataFrame?

    #     Examples
    #     --------
    #     >>> ...
    #     """

    ################################################################
    # Drawing functions
    #
    ################################################################

    # TODO: write grid draw code in separate module
    def draw(
        self,
        shape: Tuple[int, int] = (1, 4),
        shared_axes: bool = False,
        idxs: Optional[Sequence[int]] = None,
        width: Optional[int] = None,
        height: Optional[int] = None,
        margin: Union[float, Tuple[int, int, int, int]] = None,
        **kwargs,
    ) -> Tuple[Canvas, Cartesian, List[Mark]]:
        """Return a toyplot drawing of a grid of ToyTrees.

        The grid spacing can be controlled with shape and margin
        options, and trees can each be on their own axis or on a
        set of shared axes (same max height dimension), which can be
        better for highlighting differences in heights.

        Parameters
        ----------
        shape: Tuple[int,int]
            A tuple of (nrows, ncolumns) for a tree grid drawing. The
            dimension of this matrix determines the number of trees
            that will be plotted.
        shared_axes (bool):
            If True then the 'height' dimension will be shared among
            all trees, otherwise each tree is scaled to fill the
            space in its grid cell.
        idxs (int):
            The indices of trees in treelist that you want to draw. By
            default the first ncols*nrows trees are drawn, but you can
            select the 10-14th tree by entering idxs=[10,11,12,13].
            The selected trees will be displayed in the grid.
        width (int):
            Width of the canvas
        height (int):
            Height of the canvas
        margin: Tuple[int,int,int,int]
            Spacing between subplots in the grid in pixel units. A
            single value or a tuple for top, right, bottom, left.
        **kwargs (dict):
            Any style arguments supported by .draw() in toytrees.

        Examples
        --------
        >>> mtre = toytree.mtree([toytree.rtree.unittree(10) for i in range(10)])
        >>> mtre.draw(shape=(2, 3), width=800, edge_widths=4)
        """
        return draw_multitree(
            mtree=self, shape=shape, shared_axes=shared_axes, idxs=idxs,
            width=width, height=height, margin=margin, **kwargs,
        )


    # def draw_cloud_tree(
    #     self,
    #     axes: Cartesian = None,
    #     fixed_order: Sequence[str] = None,
    #     jitter: float = 0.0,
    #     idxs: Optional[Sequence[int]] = None,
    #     **kwargs,
    # ):
    #     """Return a cloud of overlapping low-opacity tree drawings.

    #     Draw multiple trees overlapping in coordinate space. The
    #     order of tip_labels is fixed in cloud trees so that trees
    #     with discordant relationships can be seen in conflict. See
    #     examples from the documentation for tips on using additional
    #     styling options to further visualize patterns among subtrees
    #     within the cloud.

    #     Parameters
    #     ----------
    #     axes: None or toyplot.coordinates.Cartesian
    #         If None then a new Canvas and Cartesian axes object is
    #         returned, otherwise if a Cartesian axes object is provided
    #         the cloudtree will be drawn on the axes.
    #     fixed_order: Sequence[str], bool, or None
    #         A list of tip names matching those in every tree in the
    #         multitree, the order of which will determine the fixed
    #         order of tips in plotted trees. If None (default) then a
    #         consensus tree is inferred and its ladderized tip order
    #         is used.
    #     jitter: float
    #         A value by which to randomly shift the baseline of tree
    #         subplots so that they do not overlap perfectly. This adds
    #         a value drawn from np.random.uniform(-jitter, jitter).
    #     idxs: None or Sequence[int]
    #         Optional select indices of which trees to draw.
    #     **kwargs:
    #         All drawing style arguments supported in the .draw()
    #         function of toytree objects are also supported by
    #         .draw_cloudtree(), most notably here, edge_style.

    #     Notes
    #     -----
    #     For style arguments that apply to tips (e.g., tip_labels) the
    #     styles will be re-ordered by fixed_order to apply to all trees
    #     correctly.
    #     """
    #     logger.warning(("kwargs", kwargs))
    #     # which trees to draw
    #     treelist = self.treelist if idxs is None else idxs

    #     # TreeStyle used for canvas, axes setup only, based on
    #     # kwargs, not .style from any subtrees.
    #     fstyle = TreeStyle()
    #     fstyle.width = kwargs.get("width", None)
    #     fstyle.height = kwargs.get("height", None)
    #     fstyle.layout = kwargs.get("layout", 'r')
    #     fstyle.padding = kwargs.get("padding", 20)
    #     fstyle.scale_bar = kwargs.get("scale_bar", False)
    #     fstyle.use_edge_lengths = kwargs.get("use_edge_lengths", True)
    #     fstyle.edge_type = kwargs.get("edge_type", 'c')
    #     fstyle.xbaseline = kwargs.get("xbaseline", 0)
    #     fstyle.ybaseline = kwargs.get("ybaseline", 0)
    #     fstyle.tip_labels = (
    #         self.treelist[0].get_tip_labels() if "tip_labels" not in kwargs
    #         else kwargs["tip_labels"]
    #     )
    #     logger.warning(("fstyle.tip_labels", fstyle.tip_labels))
    #     fstyle.edge_style.stroke_opacity = 1 / len(treelist)
    #     if "edge_style" in kwargs:
    #         for key, val in kwargs["edge_style"]:
    #             setattr(fstyle, key, val)
    #     logger.warning(("fstyle.edge_style", fstyle.edge_style))

    #     # get fixed order of tips from consensus tree if not provided.
    #     if fixed_order is None:
    #         fixed_order = (
    #             MultiTree(self.treelist)
    #             .get_consensus_tree()
    #             .get_tip_labels()
    #         )
    #     logger.warning(("fixed_order", fixed_order))

    #     # require fixed_order to match ntips
    #     ntips = self.treelist[0].ntips
    #     assert len(fixed_order) == ntips, (
    #         f"fixed_order arg (len={len(fixed_order)}) "
    #         f"must be the same length as ntips (len={ntips}). "
    #         f"You entered: {fixed_order}")

    #     # draw trees sequentially to get a list of ToyTreeMark objects
    #     marks = []
    #     for tidx, tree in enumerate(treelist):

    #         # make copy of fstyle
    #         tmpstyle = fstyle.copy()

    #         # add jitter to tip coordinates
    #         if jitter:
    #             if fstyle.layout in "rl":
    #                 tmpstyle.ybaseline = fstyle.ybaseline + np.random.uniform(-jitter, jitter)
    #             else:
    #                 tmpstyle.xbaseline = fstyle.xbaseline + np.random.uniform(-jitter, jitter)

    #         # set the edge stroke-opacity
    #         if tree.style.edge_style.stroke_opacity is not None:
    #             tmpstyle.edge_style.stroke_opacity

    #         # plot tip labels by reordering those of tree tidx=0
    #         if not tidx:
    #             tmpstyle.tip_labels = fstyle.tip_labels
    #         else:
    #             tmpstyle.tip_labels = False

    #         # add mark to axes
    #         kwargs = tree_style_to_css_dict(tmpstyle)
    #         logger.warning(("kwargs", kwargs))
    #         if not tidx:
    #             canvas, axes, mark = tree.draw(axes=axes, fixed_order=fixed_order, **kwargs)
    #         else:
    #             _, _, mark = tree.draw(axes=axes, fixed_order=fixed_order, **kwargs)
    #         marks.append(mark)

    #     # get shared tree styles.
    #     return canvas, axes, marks


    def draw_cloud_tree(
        self,
        axes: Cartesian = None,
        fixed_order: Union[Sequence[str], bool] = None,
        jitter: float = 0.0,
        idxs: Optional[Sequence[int]] = None,
        interior_algorithm: int = 1,
        **kwargs,
    ):
        """...
        """
        kwargs["jitter"] = jitter
        kwargs["axes"] = axes
        kwargs["idxs"] = idxs
        kwargs["tree_style"] = None
        kwargs["fixed_order"] = fixed_order
        kwargs["interior_algorithm"] = interior_algorithm
        kwargs["kwargs"] = {}
        marks = draw_cloudtree(self, **kwargs)

        # get or create axes and canvas
        canvas, axes = get_canvas_and_axes(
            axes, marks[0],
            kwargs.get("width"), kwargs.get("height"),
        )

        # add marks to axes
        for mark in marks:
            axes.add_mark(mark)

        # style axes
        # scale bar was not allowed on individual trees. Add scale
        # bar for the tallest tree in the bunch
        if kwargs.get("scale_bar", False):
            tree = max(self, key=lambda x: x.treenode.height)
            tree.annotate.add_axes_scale_bar(axes, ymax=tree.treenode.height)
        else:
            if canvas is not None:
                axes.x.show = False
                axes.y.show = False

        return canvas, axes, marks


    def reset_tree_styles(self):
        """Set the .style to default for all ToyTrees in treelist."""
        for tre in self.treelist:
            tre.style = TreeStyle()

    def get_tip_labels(self) -> Sequence[str]:
        """Return ordered tip labels from the first tree in treelist.

        Because a MultiTree contains many trees there are many possible
        orderings of the tip Node labels. For the purposes of plotting,
        it is sometimes desirable to fetch the tip names before plotting
        so that they can be modified, and then supplied as an argument
        to a plotting function, such as `draw_cloud_tree`.

        tree plot order for the *list of tree*, ...
        starting from the zero axis. If all trees in the treelist do not
        share the same set of tips then this will return an error message.

        If fixed_order is a user entered list then names are returned in that
        order. If fixed_order was True then the consensus tree order is
        returned. If fixed_order was None or False then the order of the first
        ToyTree in .treelist is returned.
        """
        if not self.all_tree_tip_labels_same():
            raise ToytreeError(
                "All trees in treelist do not share the same set of tips")
        return self.treelist[0].get_tip_labels()


if __name__ == "__main__":

    import toytree
    mtree = toytree.mtree([toytree.rtree.unittree(10) for i in range(10)])
    # print(mtree)
    # c, a, m = mtree.draw(ts='c', shape=(2, 3))
    # toytree.utils.show(c)
    # for tree in mtree:
    #     print(repr(tree))
    # print(mtree.get_unique_topologies())

    # c, a, m = mtree.draw_cloud_tree()
    # toytree.utils.show(c)
    mtree[1][-1].children[0]._dist = 100
    mtree[1]._update()
    c, a, m = mtree.draw(scale_bar=True, node_sizes=0, shared_axes=True, layout='d')
    toytree.utils.show(c, tmpdir="~")



#!/usr/bin/env python

"""ToyColor is a superclass of numpy array w/ the toyplot.color.dtype.

It adds several functions for easily converting between color types, 
and for automatically parsing flexible input types.

Note
----
If the color value being parsed can be either a single color or a
container object with multiple colors then the `color_expander()`
function should be used to parse it. That func will return either a
single ToyColor or List[ToyColor].
>>> ToyColor.color_expander(['red', 'blue'])
>>> # [ToyColor((1,0,0,1), ToyColor((0,0,1,1))]
"""

from __future__ import annotations
from typing import List, Any
import numpy as np
from loguru import logger
import toyplot.color
from toytree.color.src.colorkit import ColorKit
from toytree.utils import ToyColorError
from toytree.utils.src.logger_setup import capture_logs

logger = logger.bind(name="toytree")
DTYPE = {
    "names": ["r", "g", "b", "a"],
    "formats": ["float64", "float64", "float64", "float64"],
}


class ToyColor(np.ndarray):
    """ToyColor is a superclass of numpy.ndarray.

    This is similar to how toyplot uses a numpy array to represent
    color, but it can init from any valid color input type and
    convert to any other.
    """
    def __new__(cls, color):
        color = ColorKit(color)
        obj = np.array(color.rgba, dtype=DTYPE).view(cls)
        return obj

    @property
    def css(self):
        """Returns a name or hex value of a color"""
        return ColorKit(self).css

    @property
    def rgba(self):
        """Returns a tuple of (r,g,b,a) as floats."""
        return ColorKit(self).rgba

    @property
    def rgb(self):
        """Returns a tuple of (r,g,b) as floats."""
        return ColorKit(self).rgb

    @property
    def colorkit(self):
        """Returns a ..."""
        return ColorKit(self)

    def get_style_str(self, stroke: bool = False, opacity: float = None) -> str:
        """Returns a color style string, e.g., 'fill=...; stroke=...'."""
        return self.colorkit.get_style_str(stroke, opacity)

    def __str__(self):
        return self.css

    def __repr__(self):
        tup = tuple(round(i, 3) for i in self.rgba)
        return f"ToyColor({tup})"

    @classmethod
    def color_expander(cls, color: Any, check_nested: bool = True) -> List[ToyColor]:
        """Return a List of ToyColor of the proper size for data points.

        Parse the input of a color based style argument to .draw(). This
        supports a wide variety of types, with ndarray being the most
        troublesome.
        """
        # first try to parse as a single color
        try:
            return ToyColor(color)
        except ToyColorError:
            pass

        # non-iterable container types from toyplot.
        # ndarray with size > 1
        if isinstance(color, np.ndarray):
            return [ToyColor(color.item(i)) for i in range(color.size)]

        # toyplot.color.Map type
        if isinstance(color, toyplot.color.Map):
            raise ToyColorError(COLORMAP_ALONE_ERROR)

        # else it must be a multi-color type object
        if check_nested:
            with capture_logs("ERROR"):
                try:
                    return [cls.color_expander(i, check_nested=False) for i in color]
                except ToyColorError:
                    pass

        raise ToyColorError(
            f"{color} ({type(color)}) is not a supported color argument.")


COLORMAP_ALONE_ERROR = """
toyplot.color.Map objects cannot be used to enter color
values directly. Instead, use the Map object to broadcast
values into an array of colors, which you can then enter
as a valid color input. Example:
>>> cmap = toyplot.color.brewer.map('BlueRed')
>>> colors = cmap.colors([0.0, 0.1, 0.2, 0.3, 0.8, 0.9])

Or, use the tuple type input syntax as a shortcut for color mapping:
>>> tree.draw(node_colors=('dist', 'BlueRed'))
"""


if __name__ == "__main__":

    import toyplot
    # tc = ToyColor((1, 0, 0.2, 0.7))
    # print(tc.colorkit.get_style_str(stroke=True))
    # print(tc)
    # print(tc.__repr_html__)

    # parsing single colors (using ColorKit under the hood)
    COLORS = [
        "RED",
        "red",
        "rgba(100.0%,50.0%,25.0%,0.500)",
        "rgb(100%,50%,25%)",
        (1.0, 0.5, 0.25, 0.5),
        (1.0, 0.5, 0.25),
        np.array((1.0, 0.5, 0.25, 0.5), dtype=toyplot.color.dtype),
        ColorKit("red"),
        toyplot.color.Palette()[0],
    ]
    for i in COLORS:
        c = ToyColor(i)
        print(f"{repr(c):<40} {str(c):<30} \t | original = {i}")

    # parsing single or multicolors (using color_expander -> ColorKit)
    COLORS = [
        "RED",
        "red",
        "rgba(100.0%,50.0%,25.0%,0.500)",
        "rgb(100%,50%,25%)",
        "#000000",
        (1.0, 0.5, 0.25, 0.5),
        (1.0, 0.5, 0.25),
        np.array((1.0, 0.5, 0.25, 0.5), dtype=toyplot.color.dtype),
        toyplot.color.Palette(),
        toyplot.color.brewer.map("BlueRed").colors([0, 2, 1]),
        list(toyplot.color.Palette()),
        ['red', 'blue', 'green'],
        ['red', toyplot.color.Palette()[0], "rgb(100%,50%,25%)"],
    ]
    for i in COLORS:
        c = ToyColor.color_expander(i)
        if isinstance(c, ToyColor):
            print(f"orig = {str(i):<50} | {repr(c)}")
        else:
            print(f"orig = {str(type(i)):<50} | len={len(c)}, {repr([c[0], '...'])}")

    # NOT_SUPPORTED = [
    #     toyplot.color.brewer.map("Pastel1"),
    # ]


#!/usr/bin/env python

"""Compute tip angles given the current layout.

The layouts differ in how tip angles should be computed. The linear
layouts are simple, but users can still optionally modify angles. The
circular layout is also pretty straightforward, but needs to be flipped
when the names cross the 0-axis. The unrooted layout tip names are
unique for each edge, requiring computation.
"""

from typing import TypeVar
import numpy as np
from loguru import logger


logger = logger.bind(name="toytree")
ToyTree = TypeVar("ToyTree")


def get_tip_labels_angles(tree: ToyTree, coords: np.ndarray) -> np.ndarray:
    """Get tip label angles given the current layout.

    For unrooted and circular layouts the best angle to project
    tip labels depends on the coordinate layout, and so it is
    easiest to re-compute for nodes after projecting.
    """
    angles = np.zeros(tree.ntips, dtype=float)
    for node in tree.traverse():
        if node.is_leaf():
            cx, cy = coords[node.idx]

            # from grandparent, the tip label spread generally improves.
            if node.up.up:
                px, py = coords[node.up.up.idx]
            else:
                px, py = coords[node.up.idx]

            # alternative unrooted project tips from origin, sometimes
            # this is better, would be nice to make this an option.
            # px, py = coords[-1]

            # get angle based on difference between nodes
            dx = cx - px
            dy = cy - py

            # sin(x) = opp / hyp; tan_(x) = opp/adj
            if dx == 0 :
                theta = np.pi / 2
            else :
                theta = np.arctan(dy / dx)
            
            if dx < 0:
                theta += np.pi
            angles[node.idx] = np.rad2deg(theta)
            # logger.info(f"arctan2 {node.name}, {np.arctan2(dy, dx)}")
            # logger.info(
            #     f"{node.name}, {dx:.2f}, {dy:.2f}, {theta:.2f}, {angles[node.idx]:.2f}")
    return angles


if __name__ == "__main__":
    pass



#!/usr/bin/env python

"""Add additional Marks to annotate toytree drawings.

Add annotations to drawings to show data or highlight sections
of a toytree drawing. This module is available as both a package-level
module as well as from ToyTree instances as an API.

Examples
--------
>>> tree = toytree.rtree.bdtree(10, seed=123)
>>> c, a, m = tree.draw(layout='d')
>>> toytree.annotate.add_node_pie_charts(axes=a, **kwargs)

>>> tree = toytree.rtree.bdtree(10, seed=123)
>>> c, a, m = tree.draw(layout='d', xbaseline=10)
>>> tree.annotate.node_pie_charts(axes=a, layout='d', xbaseline=10)

>>> # other idea, pass the tree Mark as an arg
>>> c, a, m = tree.draw(layout='d', xbaseline=10)
>>> toytree.annotate.node_pie_charts(mark=m)

Functions
---------
- add_node_markers()
- add_node_labels()
- add_node_pie_charts()

- add_edge_markers()
- add_edge_labels()
- add_edge_pie_charts()
- add_edge_root()

- add_scale_bar()
- add_confidence_intervals()

- add_tip_markers()
- add_tip_labels()
- add_tip_heatmap()

- add_image_markers(xpos, ypos, image[s], size, **style)

- radial_...
"""

# from toytree.annotate.src.node_pie_charts import draw_node_pie_charts
from toytree.annotate.src.add_edge_markers import *
from toytree.annotate.src.add_node_markers import *
from toytree.annotate.src.add_tip_markers import *
from toytree.annotate.src.add_axes_box_outline import *
from toytree.annotate.src.add_scale_bar import *

# ... edge_labels
# ... node_labels
# from .confidence_intervals import draw_node_confidence_intervals



#!/usr/bin/env python

"""Write trees to newick or nexus formats.

Performs a `tree_reduce` operation to recursively condense the
tree structure into a nested string in parenthetical format, i.e.,
NEWICK format. This can optionally be wrapped with additional
information as a NEXUS formatted file.

References
----------
- https://gist.github.com/Ad115/fbee7a7a85935888d383f05b7f99e956

"""

from typing import Optional, Tuple, Sequence, Callable
import numpy as np
from loguru import logger
from toytree.core import ToyTree, Node
from toytree.utils import ToytreeError
from toytree.core.apis import add_toytree_method

DISALLOWED_FEATURES = set(['idx', 'dist', 'up', 'children'])  # 'support', 'height'])
logger = logger.bind(name="toytree")


def get_float_formatter(formatting_str: str) -> Callable:
    """Return func to format string using % or {} notation.
    """
    if formatting_str is None:
        def formatter(x):
            return ""
    elif "%" in formatting_str:
        def formatter(x):
            return formatting_str % x
    elif formatting_str.startswith("{"):
        def formatter(x):
            return formatting_str.format(x)
    else:
        raise ValueError(
            "'features_formatter' is not a proper Python formatting str: "
            f"{formatting_str}")
    return formatter


def get_feature_string(
    node: Node,
    features: Sequence[str],
    features_prefix: str,
    features_delim: str,
    features_assignment: str,
    features_formatter: str,
) -> str:
    """Return a string of commented features inside square brackets.

    Intended to handle formatting/serializeation of flexible object
    types that could be assigned to Nodes.
    """
    if not features:
        return ""
    pairs = []

    # get float formatting function
    formatter = get_float_formatter(features_formatter)

    # check each feature
    for feature in features:
        # get value or None if Node does not have the feature
        value = getattr(node, feature, None)

        # if no value then do not write it
        if value is None:
            continue

        # try to convert to float and float format the value
        try:
            value = float(value)
            if np.isnan(value):
                value = ""
            else:
                value = formatter(value)
        except (TypeError, ValueError):
            value = str(value)

        # store if a value exists
        if value:
            pairs.append(features_assignment.join((feature, value)))
    feature_str = features_delim.join(pairs)
    if not feature_str:
        return ""
    return f"[{features_prefix}{feature_str}]"


def node_to_newick(
    node: Node,
    children: Tuple[Node],
    dist_formatter: str = "%.12g",
    internal_labels: Optional[str] = "support",
    internal_labels_formatter: Optional[str] = "%.12g",
    node_features: Optional[Sequence[str]] = None,
    edge_features: Optional[Sequence[str]] = None,
    features_prefix: str = "&",
    features_delim: str = ",",
    features_assignment: str = "=",
    features_formatter: str = "%.12g",
    names_as_ints: bool = False,
) -> str:
    """Reduce function used in tree_reduce"""
    # format the comment feature string for extra features
    node_feature_str = get_feature_string(
        node,
        node_features,
        features_prefix,
        features_delim,
        features_assignment,
        features_formatter,
    )
    edge_feature_str = get_feature_string(
        node,
        edge_features,
        features_prefix,
        features_delim,
        features_assignment,
        features_formatter,
    )

    # format the dist values (edge lengths) as strings
    dist_formatter = get_float_formatter(dist_formatter)
    dist = dist_formatter(node.dist)
    dist = dist if not dist else f":{dist}"

    # format the internal label feature (usually support, name, or "")
    label_formatter = get_float_formatter(internal_labels_formatter)
    if internal_labels is None:
        internal = ""
    else:
        internal = getattr(node, internal_labels, "")
        if internal is None:
            internal = ""
        else:
            try:
                internal = float(internal)
                if np.isnan(internal):
                    internal = ""
                else:
                    internal = label_formatter(internal)
            except (TypeError, ValueError):
                internal = str(internal)

    # tip node write N[meta]:E[emeta] if dist else N[nmeta]
    if node.is_leaf():
        if names_as_ints:
            if dist:
                return f"{node.idx}{node_feature_str}{dist}{edge_feature_str}"
            return f"{node.idx}{node_feature_str}"
        if dist:
            return f"{node.name}{node_feature_str}{dist}{edge_feature_str}"
        return f"{node.name}{node_feature_str}"

    # root node write N[nmeta] unless the root has dist then N[meta]:E[emeta]
    if node.is_root():
        if node.dist:
            return f"({','.join(children)}){internal}{node_feature_str}{dist}{edge_feature_str}"
        return f"({','.join(children)}){internal}{node_feature_str}"

    # other internal nodes write N[nmeta]:E[emeta] if dist else N[nmeta]
    if dist:
        return f"({','.join(children)}){internal}{node_feature_str}{dist}{edge_feature_str}"
    return f"({','.join(children)}){internal}{node_feature_str}"


def tree_reduce(
    node: Node,
    dist_formatter: str = "%.12g",
    internal_labels: Optional[str] = "support",
    internal_labels_formatter: Optional[str] = "%.12g",
    node_features: Optional[Sequence[str]] = None,
    edge_features: Optional[Sequence[str]] = None,
    features_prefix: str = "&",
    features_delim: str = ",",
    features_assignment: str = "=",
    features_formatter: str = "%.12g",
    names_as_ints: bool = False,
) -> str:
    """Return newick string of ToyTree.

    Recursive function to get formatted newick string of tree data.
    See `write_newick` for docstring.
    """
    args = [
        dist_formatter, internal_labels, internal_labels_formatter,
        node_features, edge_features,
        features_prefix, features_delim, features_assignment,
        features_formatter,
        names_as_ints,
    ]
    reduced_children = [tree_reduce(child, *args) for child in node.children]
    newick = node_to_newick(node, reduced_children, *args)
    # logger.warning([node, newick])
    return newick


def wrap_nexus(tree: ToyTree, newick: str) -> str:
    """Wrap a newick string into NEXUS format.

    """
    nexus = "#NEXUS\n"
    nexus += "begin trees;\n"

    # add translation using idx
    nexus += "    translate\n"
    for node in tree[:tree.ntips]:
        nexus += f"        {node.idx} {node.name},\n"
    nexus += "    ;\n"

    # get rooting info
    rooted = "R" if tree.is_rooted() else "U"

    # write tree
    nexus += f"    tree 0 = [&{rooted}] {newick}\n"
    nexus += "end;"
    return nexus


@add_toytree_method(ToyTree)
def write(
    tree: ToyTree,
    path: Optional[str] = None,
    dist_formatter: Optional[str] = "%.12g",
    internal_labels: Optional[str] = "support",
    internal_labels_formatter: Optional[str] = "%.12g",
    features: Optional[Sequence[str]] = None,
    features_prefix: str = "&",
    features_delim: str = ",",
    features_assignment: str = "=",
    features_formatter: str = "%.12g",
    nexus: bool = False,
    **kwargs,
) -> Optional[str]:
    """Write tree to newick string and return or write to filepath.

    The newick string can be formatted in several ways. The default
    will include dist values (edge lengths) and support values as
    internal node labels. The edge lengths can be suppressed by
    setting `dist_formatter=None`, and internal node labels can be
    similarly suppressed, or set to store a different feature, such
    as internal node names. Additional features can be stored in the
    node comment blocks in extended-newick-format (NHX-like) by using
    the "features" arguments (see examples).

    Parameters
    ----------
    tree: ToyTree
        A ToyTree instance to write as a newick string.
    path: str or None
        A filepath to write to file, or None to return newick string.
    dist_formatter: str or None
        A formatting string to format float dist values (edge lengths).
        Default is "%.12g". If None edge lengths are excluded.
    internal_labels: str or None
        A feature to write as internal node labels. None suppresses
        internal labels. Default is 'support', 'name' is also commonly
        used here, but any feature name can be selected.
    internal_labels_formatter: str or None
        A formatting string to format internal labels as floats.
    features: List[str]
        A list of additional features to write in the newick comment
        block. For example, features=["height"] will save heights.
    features_prefix: str
        A prefix character written to the start of newick comment
        blocks. Typical values are "&" (default) or "&&NHX:".
    features_delim: str
        A character used to delimit features in the newick comment
        block. Default is ",".
    features_assignment: str
        A character used to separate feature keys and values. Default
        is "=".
    features_formatter: str or None
        A formatting string used for float feature metadata. Default
        is "%.12g".
    nexus: bool
        If True the tree data is nested in a "trees" block, names are
        translated to ints, and a #NEXUS header is included.

    See Also
    --------
    `ToyTree.write`
        This function is available from ToyTree objects as `.write`.

    Examples
    --------
    >>> # generate a tree with additional metadata
    >>> tree = toytree.rtree.baltree(ntips=4)
    >>> tree.set_node_data("name", {4: "A", 5: "B", 5: "C"}, inplace=True)
    >>> tree.set_node_data("support", {4: 100, 5: 90}, inplace=True)
    >>> tree.set_node_data("X", default=10, inplace=True)

    >>> # write tree to file path
    >>> tree.write(path="/tmp/test.nwk")

    >>> # return tree as str with various formatting options.
    >>> tree.write()
    >>> '((r0:0.5,r1:0.5)100:0.5,(r2:0.5,r3:0.5)90:0.5);'
    >>> tree.write(dist_formatter=None)
    >>> '((r0,r1)100,(r2,r3)90);'
    >>> tree.write(internal_labels=None)
    >>> '((r0:0.5,r1:0.5):0.5,(r2:0.5,r3:0.5):0.5);'
    >>> tree.write(internal_labels="name")
    >>> '((r0:0.5,r1:0.5)A:0.5,(r2:0.5,r3:0.5)C:0.5);'
    >>> tree.write(features=["X"])
    >>> '((r0[&X=10]:0.5,r1[&X=10]:0.5)100[&X=10]:0.5,(r2[&X=10]:...
    """
    if kwargs:
        logger.warning(f"Deprecated args to write(): {list(kwargs)}. See docs.")

    # separate node and edge features
    if features is None:
        features = []
    features = [features] if isinstance(features, str) else features
    features = set(features) - DISALLOWED_FEATURES
    bad_features = features - set(tree.features)
    if bad_features:
        raise ToytreeError(
            f"Cannot write features not present in tree: {bad_features}")
    node_features = {i for i in features if i not in tree.edge_features}
    edge_features = features - node_features

    # build newick string from recursive
    newick = tree_reduce(
        tree.treenode,
        dist_formatter,
        internal_labels,
        internal_labels_formatter,
        node_features,
        edge_features,
        features_prefix,
        features_delim,
        features_assignment,
        features_formatter,
        names_as_ints=nexus,
    ) + ";"

    # optionally wrap as nexus
    treestr = wrap_nexus(tree, newick) if nexus else newick

    # write to file or return
    if path is not None:
        with open(path, 'w', encoding="utf-8") as out:
            out.write(treestr)
            return None
    return treestr


if __name__ == "__main__":

    import toytree
    NWK = "((a:3[&state=1],b:3[&state=1])D:1[&state=1],c:4[&state=2])E:1[&state=1];"
    TREE = toytree.tree(NWK)
    # TREE = TREE.set_node_data("support", default=100)

    print(write(TREE))
    print(write(TREE, dist_formatter=None))
    print(write(TREE, internal_labels=None))
    print(write(TREE, internal_labels="name"))
    print(write(TREE, features=["state"]))
    # print(write_nexus(TREE, features=["state"]))

    TREE.set_node_data('test', {'a': 7.8, 4: 5}, inplace=True)
    TREE.set_node_data('support', default=100, inplace=True)
    # print(TREE.edge_features)
    # print(write(TREE, internal_labels=None, features=["support", "test"], nexus=True))
    print(write(TREE, dist_formatter=None, features=["name", "support"]))


    TREE.write()



#!/usr/bin/env python

"""Unrooted layout.

"""

from typing import TypeVar, Tuple
import numpy as np
from loguru import logger
from toytree.layout import BaseLayout, get_tip_labels_angles

logger = logger.bind(name="toytree")
ToyTree = TypeVar("ToyTree")


class UnrootedLayout(BaseLayout):
    """Layout for unrooted tree projection: "unrooted", "u1", "u2"
    """
    def run(self):
        """Fills the .coords array with x, y values."""

        # xbaseline, ybaseline to set origin.
        self.coords = equal_daylight_algorithm(
            tree=self.tree,
            max_iter=50,
            use_edge_lengths=self.style.use_edge_lengths)
        self.tcoords = self.coords[:self.tree.ntips, :].copy()
        self.angles = get_tip_labels_angles(self.tree, self.coords)
        self.style_overwrite()

    def style_overwrite(self):
        """Overwrite styles that are not allowed in unrooted layout
        of which should have different defaults.
        """
        self.style.tip_labels_align = False
        # self.style.tip_labels_style._toyplot_anchor_shift = 0
        # self.style.tip_labels_style.text_anchor = "middle"
        self.style.edge_type = 'c'


#####################################################
# UNROOTED LAYOUT UTILITIES
#####################################################


def rotate_arr(
    points: np.ndarray,
    origin: Tuple[float, float] = (0, 0),
    degrees: float = 0,
) -> np.ndarray:
    """Return an array of rotated Node coordinates."""
    angle = np.deg2rad(degrees)
    rotation_matrix = np.array([
        [np.cos(angle), -np.sin(angle)],
        [np.sin(angle), np.cos(angle)],
    ])
    origin = np.atleast_2d(origin)
    point = np.atleast_2d(points)
    return np.squeeze((rotation_matrix @ (point.T - origin.T) + origin.T).T)


def equal_daylight_algorithm(
    tree: ToyTree,
    max_iter: int = 1,
    min_delta: float = 1,
    use_edge_lengths: bool = True,
) -> float:
    """Return coordinates for unrooted layout under the eda algorithm.

    This algorithm equalizes the sizes of angular gaps between
    subtrees. As Felsenstein said, the result is "outstanding".
    The description however is not very detailed, so there is room
    for interpretation.

    I use a levelorder traversal but could it seems that postorder
    can yield better results sometimes, but also much worse sometimes.
    If the eda layout is too different from the eaa layout it will
    be rejected.

    Notes
    -----
    We could estimate the length that tip names will extend past the
    end of tips when calculating daylight. This would ensure that tip
    names to do not crossover, not just that edges do not cross over.
    This would only be necessary if tip names were text-align=start,
    however, but they are usually drawn all with angle=0, so the idea
    is moot I suppose.

    References
    ----------
    - Felsenstein 2004, page 582 (and see Figure 34.6).
    """
    # get the equal angle algorithm tree as a starting tree.
    coords = equal_angle_algorithm(tree, use_edge_lengths=use_edge_lengths)

    # return equal-angles for <= 4-taxon trees
    if (tree.ntips < 5) | (max_iter == 0):
        return coords

    # a set with all Nodes except the root
    # nodes = set(tree.get_nodes()[:-1])
    nodes = set(tree.get_nodes())

    # for Y internal nodes we expect an average rotation of X, thus
    # any solution with more delta than 3 * Y * X is almost surely
    # worse than the equal-angles layout and should be discarded.
    n_internal = tree.nnodes - tree.ntips - 1
    min_angles = n_internal * 3
    avg_change = 30
    max_change = (min_angles * avg_change) / 3  # .5

    # Perform multiple passes through the tree stopping when either
    # the improvement falls below a threshold, max_iters is reached,
    # or the iteration results in more changes than a previous one.
    sum_deltas = []  # list of sum change in angles each iter
    full_circle = None  # record whether full circle encountered.
    niter = 0
    while 1:
        # create a copy of coords to modify
        icoords = coords.copy()

        # record the sum of changed angles this iter
        sum_delta = 0

        # visit each internal node in levelorder and rotate other nodes
        # relative to the position of this one. Skip root and leaves,
        # they only rotate relative to internals.
        for fnode in tree.traverse("levelorder"):

            # only visit internal non-root nodes.
            # if fnode.is_root():
                # continue
            if fnode.is_leaf():
                continue

            # get current focal node coordinates
            pos = icoords[fnode.idx]

            # get the 3 or more subtrees connected to this vertex
            full_set = nodes - {fnode}
            subsets = [set(i.get_descendants()) for i in fnode.children]
            other_set = full_set - set.union(*subsets)
            if other_set:
                subsets.append(full_set - set.union(*subsets))
            subsets = dict(enumerate(subsets))

            # get shaded sectors by subtrees.
            shade = {}
            last = 0
            for sidx, subset in subsets.items():
                rads = []
                for node in subset:

                    # get position of this node relative to focal node
                    npos = icoords[node.idx]
                    delta_x = npos[0] - pos[0]
                    delta_y = npos[1] - pos[1]

                    # get start,end radians of sector where 0 is down
                    rad = np.arctan2(delta_y, delta_x) + np.pi / 2
                    # flip 4th quad negative to positive
                    rad = (2 * np.pi) + rad if rad < 0 else rad
                    # save 1st quad as >2pi
                    rad = (2 * np.pi) + rad if rad < last else rad
                    rads.append(rad)

                # store the sector of each shaded region
                shade[sidx] = (min(rads), max(rads))
                last = max(rads)

            # get the order in which shaded regions rotate around node
            skeys = sorted(shade, key=lambda x: abs(shade[x][0]))
            gaps = [(skeys[i], skeys[i + 1]) for i in range(len(skeys) - 1)]
            gaps += [(skeys[-1], skeys[0])]

            # get daylight as sectors between ordered shaded regions
            # e.g., (0, 1), (1, 2), ...
            light = {}
            for gap in gaps:
                last, this = shade[gap[0]][1], shade[gap[1]][0]
                if last > this:
                    this = 2 * np.pi + this
                light[gap] = np.rad2deg(this - last)

                # debugging daylight for each gap
                # logger.debug(
                #     f"{fnode.idx} daylight: {gap}: size={light[gap]:.1f} "
                #     f"-- last={np.rad2deg(last):.1f}, this={np.rad2deg(this):.1f}"
                # )

            # if subtree angles sum to more than a full circle it is
            # impossible to find a positive angle between subtrees and
            # they will overlap. So we stop the iteration and return
            # last set of coordinates. This will not occur when starting
            # from the equal-angles layout, but can on any subsequent iters.
            if (shade[skeys[-1]][1] - (2 * np.pi)) > shade[skeys[0]][0]:
                full_circle = fnode.idx

            # change angles
            optimal = sum(light.values()) / len(light)      # 56
            rotated = [0]
            for gap in gaps[:-1]:
                # how far the gap[1] needs to be rotated
                gap_to_next = light[gap] + rotated[-1]      # 76 +  0 = 76, 63 + 20
                angle_to_rotate = gap_to_next - optimal     # 76 - 56 = 20, 83 - 56
                rotated.append(angle_to_rotate)             # [0, 20,],

                # rotation = angle_to_rotate
                nodes_to_rotate = subsets[gap[1]]           # [13,4,5],
                idxs = [i.idx for i in nodes_to_rotate]
                icoords[idxs, :] = rotate_arr(
                    points=icoords[idxs],
                    origin=pos,
                    degrees=-angle_to_rotate,
                )
                # debugging rotation applied to each Node
                # logger.debug(
                #     f"node {fnode.idx} rotating {len(nodes_to_rotate)} nodes by "
                #     f"{angle_to_rotate:.2f} deg to get optimal daylight {optimal:.1f} deg.")

            # record how much change has happened
            avg_delta = np.mean([abs(i) for i in rotated[1:]])
            sum_delta += avg_delta

        # iteration finished. Should we accept this change to coordinates?
        # logger.debug(
        #     f"sum angles change={sum(sum_deltas):.1f}, "
        #     f"angles_change_this_iter={sum_delta:.1f}")

        # causes to not accept the proposed coordinate change.
        if full_circle is not None:
            # logger.debug(f"stopping at {niter} iters because subtree angles sum to > full circle @ node {full_circle}")
            break
        if sum_delta > max_change:
            # logger.debug(f"stopping at {niter} iters because bad solution ({sum_delta} > {max_change}).")
            break
        if sum_deltas:
            last = sum_deltas[-1]
            if sum_delta > (last + last * .1):
                # logger.debug(f"stopping at {niter} iters because encountered a worse solution.")
                break

        # accept the coordinates change
        niter += 1
        sum_deltas.append(sum_delta)
        coords = icoords

        if niter == max_iter:
            # logger.debug(f"stopping after {niter} iters because max_iter reached.")
            break
        if sum_delta < min_delta:
            # logger.debug(f"stopping after {niter} iters because delta <= min_delta.")
            break

    return coords


def equal_angle_algorithm(
    tree: ToyTree,
    use_edge_lengths: bool = True,
) -> float:
    """Return coordinates for unrooted layout under the 'eaa' algorithm.

    Assign the root node a sector from 0-360 degrees, and divide each
    descendent node into subsectors with size weighted by their n
    descendants.

    Parameters
    ----------
    tree: ToyTree
        A Toytree from which to build coordinates.
    use_edge_lengths: bool
        If False then all dists are set to 1.

    References
    ----------
    - Felsenstein (2004), page 578.
    """
    coords = np.zeros(shape=(tree.nnodes, 2))

    # if tree is rooted then use root Node as the central vertex.
    ntips = tree.ntips
    radians_per_tip = 2 * np.pi / ntips

    # store {Node: value}
    radian_sums = {}
    sectors = {}

    # record the sum of sector area for each Node as its N
    # descendants * the radians per tip.
    for node in tree.traverse("postorder"):
        if node.is_leaf():
            # node.radian_sum = radians_per_tip
            radian_sums[node] = radians_per_tip
        else:
            # node.radian_sum = sum(i.radian_sum for i in node.children)
            radian_sums[node] = sum(radian_sums[i] for i in node.children)

    # assign radian sectors in levelorder.
    for node in tree.traverse("levelorder"):
        if node.is_root():
            coords[node.idx] = (0, 0)
            sectors[node] = [0, 2 * np.pi]
        else:
            cohort = node.up.children
            idx = cohort.index(node)
            if not idx:
                start = sectors[node.up][0]
            else:
                start = sectors[cohort[idx - 1]][1]
            sectors[node] = [start, start + radian_sums[node]]
            mid = sum(sectors[node]) / 2.

            # geometry relative to parent position and angle
            parent_pos = coords[node.up.idx]
            hypo = node.dist if use_edge_lengths else 1.
            newx = parent_pos[0] + (hypo * np.sin(mid))
            newy = parent_pos[1] - (hypo * np.cos(mid))
            coords[node.idx] = (newx, newy)
    return coords


# NOT CURRENTLY USED
# def get_subtrees(tree: ToyTree, node: Node) -> Sequence[Set[Node]]:
#     """Return subtrees connected to a Node.

#     If this Node was removed from an unrooted tree it would induce
#     three or more subtrees. This returns the sets of Nodes in each
#     subtree. This is used internally for unrooted tree layouts.
#     """
#     nodes = set(tree.get_nodes()) - {node}
#     subsets = [set(i.get_descendants()) for i in node.children]
#     subsets.append(nodes - set.union(*subsets))
#     return subsets


if __name__ == "__main__":

    import toytree
    # Felsenstein book example
    NWK = "(((((((A:4,B:4):6,C:5):8,D:6):3,E:21):10,((F:4,G:12):14,H:8):13):13,((I:5,J:2):30,(K:11,L:11):2):17):4,M:56);"
    # NWK = "(((E,F),(G, H)),((C,D),(B,(I,J)),A));"
    TRE = toytree.tree(NWK)
    # TRE = toytree.rtree.bdtree(20, b=0.5, d=0.5, seed=123)
    # TRE = TRE.unroot()

    toytree.set_log_level("INFO")
    coords1 = equal_angle_algorithm(TRE, False)
    coords2 = equal_daylight_algorithm(TRE, max_iter=10)

    import toyplot

    canvas1 = toyplot.Canvas(400, 400)
    axes = canvas1.cartesian()
    axes.graph(
        TRE.get_edges(),#.values,
        vcoordinates=coords1,
        # vsize=16,
        estyle={"stroke-width": 2},
    )
    canvas2 = toyplot.Canvas(500, 400)
    axes = canvas2.cartesian()
    axes.graph(
        TRE.get_edges(),#.values,
        vcoordinates=coords2,
        #vsize=16,
        estyle={"stroke-width": 2},
        ecolor='black',
    )

    c, a, m = TRE.draw(node_labels="idx")
    toytree.utils.show([canvas1, canvas2, c], new=False)
    # toyplot.browser.show([canvas1, canvas2, c])
    # print(equal_daylight_algorithm(tre))

    # lay = Layout(
    #     tre,
    #     layout='c',
    #     fixed_order=tre.get_tip_labels()[::-1],
    #     fixed_position=None,
    #     xbaseline=10,
    # )
    # print(lay.coords)

    # lay = Layout(
    #     tre,
    #     layout='c0-180',
    #     fixed_order=tre.get_tip_labels()[::-1],
    #     fixed_position=None,
    #     xbaseline=10,
    # )
    # print(lay.coords)

    # # unrooted layout
    # lay = Layout(
    #     tre,
    #     layout="*",
    #     fixed_order=tre.get_tip_labels()[::-1],
    #     fixed_position=None,
    #     xbaseline=10,
    # )
    # print(lay.coords)




#!/usr/bin/env python

"""Function to set data to all Node objects in a tree.

"""

from typing import Union, Mapping, Sequence, Any, TypeVar
from loguru import logger
import pandas as pd
from toytree import ToyTree, Node
from toytree.core.apis import add_toytree_method
from toytree.data._src.expand_node_mapping import expand_node_mapping
from toytree.utils import ToytreeError


Query = TypeVar("Query", int, str, Node)
logger = logger.bind(name="toytree")


INVALID_SET_NODE_DATA_TYPE = """
Invalid 'data' arg to set_node_data(). Must be either a Mapping or a Series
of length nnodes. This appears to be a Series of len={} while nnodes={}.\
"""

# one of the following types:
# - Mapping[Query: Any], e.g., `data={0: 3, 1: 4}`
# - Series[Any] in Node idx order w/ len=nnodes, e.g., `data=range(tree.nnodes)`
# """


@add_toytree_method(ToyTree)
def set_node_data(
    tree: ToyTree,
    feature: str,
    data: Union[Mapping[Query, Any], Sequence[Any]] = None,
    default: Any = None,
    inherit: bool = False,
    inplace: bool = False,
) -> ToyTree:
    """Create or modify features (data) set to Nodes in a ToyTree.

    Features can be set on all or only some Nodes. In the latter case
    a value for Nodes with missing features can be imputed when you
    call the function `get_node_data`. Some features used internally
    are protected from modification (e.g., idx, up, children), but
    other base features such as name, dist, height, and support can be
    modified, and any new named feature can be created or modified.
    Note that a modified tree copy is returned unless `inplace=True`.

    Parameters
    -----------
    feature: str
        The name of the node attribute to create or modify.
    data: Mapping or Series
        A mapping of {Node Query: value} for one or more Nodes, or a
        Series[value] for all Nodes in Node idx traversal order.
    default: Any
        Optionally set a default value to all Nodes not included in
        data mapping.
    inherit: bool
        If inherit is True then feature values entered as a mapping to
        internal Nodes will also be applied to their descendant Nodes,
        allowing for easy assignment of values to clades.
    inplace: bool
        If True the tree data is modified inplace and returned, else
        the original tree data is unchanged and a copy is returned.

    Data can be set to Nodes by entering `data` as either a Mapping or
    as a Series of values. These two options differ slightly and are
    described below.

    Mapping: Values can be set by providing a 'mapping' (e.g., dict or
    pd.Series) with Node queries (i.e., Node, idx label, str name, or
    name regular expression) mapped to feature values for any number
    of Nodes. The 'default' option can be used to set a value to all
    Nodes not specified in the mapping. Otherwise, any Nodes not in
    the data arg will not have values set, and will maintain either no
    attribute for this feature, or any existing attribute.

    Series: Alternatively, data can be set to Nodes on a tree as a
    Series (e.g., List, np.ndarray) of values instead of a Mapping.
    In this case the Series must be len=nnodes, and the values should
    be listed in Node idx traversal order. The 'default' arg value will
    still apply to any Nodes given a value of None.

    See Also: `get_node_data`

    Examples
    --------
    Set data to Nodes using a variety of Node query dicts
    >>> tree = toytree.rtree.unittree(ntips=10)
    >>> new = tree.set_node_data(feature="Ne", default=5000)
    >>> new = tree.set_node_data(feature="Ne", data={0:1e5, 1:1e6, 2:1e3})
    >>> new = tree.set_node_data(feature="Ne", data={0:1e5, 1:1e6}, default=5000)
    >>> new = tree.set_node_data(feature="Ne", data={'r0':1e5, 'r1':1e6})
    >>> new = tree.set_node_data(feature="Ne", data={'~r[0-5]+': 1e5})
    >>> new = tree.set_node_data("state", {10: "A", 11: "B"}, inherit=True")

    or, set data to Nodes as a Series
    >>> new = tree.set_node_data("X", range(tree.nnodes))
    """
    # immutable; do not allow modifying topology attributes
    if feature in ["idx", "up", "children"]:
        raise ToytreeError(
            f"cannot modify '{feature}' feature because it affects the "
            "tree topology. To modify topology see `toytree.mod` "
            "subpackage functions.")

    # try to convert data to a Dict[Node, Any] if it is a dict-like
    if data is None:
        mapping = {}
    else:
        try:
            mapping = dict(data)
            mapping = expand_node_mapping(tree, mapping)
        except (ValueError, TypeError):
            if not len(data) == tree.nnodes:
                msg = INVALID_SET_NODE_DATA_TYPE.format(len(data), tree.nnodes)
                logger.error(msg)
                raise ToytreeError(msg)
            mapping = dict(zip(range(tree.nnodes), data))
            mapping = expand_node_mapping(tree, mapping)

        # report non-TypeError mapping exception
        except Exception as exc:
            raise ToytreeError(INVALID_SET_NODE_DATA_TYPE) from exc

    # make a copy of ToyTree to return
    tree = tree if inplace else tree.copy()

    # copy mapping to {int: value} and optionally map to descendants
    ndict = {}

    # sort {Node: feat} keys into reverse idx order for inherit=True
    nkeys = sorted(mapping, reverse=True, key=lambda x: x._idx)

    # fill new dict w/ inherited values
    for node in nkeys:
        value = mapping[node]
        ndict[node._idx] = value
        if inherit is True:
            for desc in node.iter_descendants():
                ndict[desc._idx] = value

    # map {Node: default} for Nodes not in ndict
    if default is not None:
        for node in tree:
            if node._idx not in ndict:
                ndict[node._idx] = default

    # special mod submodule method for height modifications
    if feature == "height":
        height_map = {i: j for (i, j) in ndict.items() if j is not None}
        return tree.mod.edges_set_node_heights(height_map, inplace=inplace)

    # dist is immutable, but allow it here, and do an update.
    if feature == "dist":
        feature = "_dist"

    # add value to Nodes as a feature. If the value can be copied,
    # e.g., a dict, array, etc., then assign copies, otherwise if
    # this object is changed it affects the value of multiple Nodes
    for nidx, value in ndict.items():
        if hasattr(value, 'copy'):
            setattr(tree[nidx], feature, value.copy())
        else:
            setattr(tree[nidx], feature, value)

    # if dist was mod'd then must call update
    if feature == "_dist":
        tree._update()
    return tree


@add_toytree_method(ToyTree)
def set_node_data_from_dataframe(
    tree: ToyTree,
    table: pd.DataFrame,
    inplace: bool = False,
) -> ToyTree:
    """Set new features on Nodes of a ToyTree from a DataFrame.

    The DataFrame should have column names corresponding to features
    that you wish to apply to Nodes of the ToyTree. The index can
    be composed of either strings that match to .name attributes
    of Nodes in the ToyTree, or can be integers, which match to the
    .idx labels of Nodes. Note: to set data to internal Nodes that
    usually do not have unique name labels you will likely need to
    use the numeric idx labels. Be aware that idx labels are
    unique to each topology, and will change if the tree topology
    is modified.

    This function parses the DataFrame and applies the function
    `set_node_data()` for each column.

    Parameters
    ----------
    table: pd.DataFrame
        A DataFrame with data to be applied to Nodes of a ToyTree.
    inplace: bool

    Returns
    -------
    A copy of the original ToyTree with node features modified.

    See Also
    --------
    :meth:`~toytree.core.tree.ToyTree.get_node_data`,
    :meth:`~toytree.core.tree.ToyTree.set_node_data`.

    Examples
    --------
    >>> tree = toytree.rtree.unittree(ntips=10)
    >>> data = pd.DataFrame({
    >>>    'trait1': np.arange(tree.nnodes),
    >>>    'trait2': np.random.randint(0, 100, tree.nnodes),
    >>> })
    >>> tree = tree.set_node_data_from_dataframe(data)
    >>> tree.get_node_data()
    """
    # make a copy of ToyTree to return
    tree = tree if inplace else tree.copy()
    for key in table.columns:
        mapping = table[key].to_dict()
        tree.set_node_data(feature=key, data=mapping, inplace=True)
    return tree


if __name__ == "__main__":

    import toytree
    import numpy as np
    toytree.set_log_level("INFO")

    tree = toytree.rtree.unittree(ntips=10)
    new_tree = set_node_data(tree, feature="height", data={0: 10, 'r1': 20, tree[3]: 50}, default=5000)    
    new_tree = set_node_data(tree, feature="Ne", default=5000)
    new_tree = set_node_data(tree, feature="Ne", data={0: 1e5, 1: 1e6, 2: 1e3})
    new_tree = set_node_data(tree, feature="Ne", data={0: 1e5, 1: 1e6}, default=5000)
    new_tree = set_node_data(tree, feature="Ne", data={'r0': 1e5, 'r1': 1e6})
    new_tree = set_node_data(tree, feature="Ne", data={'r0': 1e5, '~r[2-5]+': 1e6})
    new_tree = set_node_data(tree, feature="Ne", data={11: 1e5}, inherit=True)
    new_tree = set_node_data(tree, feature="Ne", data={11: 1e5})
    new_tree = set_node_data(tree, feature="Ne", data=range(tree.nnodes))
    new_tree = set_node_data(tree, "Ne", {11: "red"}, default="blue", inherit=True)
    new_tree = set_node_data(tree, "Ne", {5: 5, 6: 6})

    # # typeerror,valuerrror
    # new_tree = set_node_data(tree, "Ne", ['a', 'b'] + [np.nan] * (tree.nnodes - 2))
    # new_tree = tree.set_node_data(
    #     feature="state",
    #     mapping={10: "A", 11: "B"},
    #     inherit=True,
    # )
    # print(new_tree.get_node_data("Ne"))

    tre = toytree.rtree.unittree(8, seed=123)
    anc = tre.get_ancestors(1, 2, 3)
    tre = tre.set_node_data("color", {i.idx: "red" for i in anc}, default="blue")
    print(tre.get_node_data())



#!/usr/bin/env python

"""Canvas setup functions for single or grids of trees."""

from typing import List
from dataclasses import dataclass, field
import numpy as np
from toytree.core import Canvas, Cartesian


# def get_linear_width_and_height(mark: MultiTreeMark) -> Tuple[int, int]:
#     """Return height and width for tree grid if not set by user."""
#     if mark.layout in "du":
#         minx = 225
#         miny = 250
#         width = mark.width if mark.width else min(750, minx * mark.ncols)
#         height = mark.height if mark.height else min(750, miny * mark.nrows)
#         return width, height

#     minx = 250
#     miny = 225
#     height = mark.height if mark.height else min(750, minx * mark.nrows)
#     width = mark.width if mark.width else min(750, miny * mark.ncols)
#     return width, height


@dataclass
class Grid:
    nrows: int
    ncols: int
    width: int
    height: int
    layout: str
    margin: int
    padding: int
    scale_bar: bool

    # attrs to be filled
    canvas: Canvas = field(default=None, init=False)
    axes: List[Cartesian] = field(init=False, default_factory=list)

    def __post_init__(self):
        self.get_canvas()
        self.get_axes()

    def get_axes(self) -> None:
        """Set .axes attribute with Cartesian and set margin.
        """
        nplots = self.nrows * self.ncols
        grid = np.arange(nplots).reshape((self.nrows, self.ncols))

        # create a separate Cartesian for each tree subplot
        for idx in range(nplots):

            # get the margin setting depending on axes on edge or mid
            if self.margin is not None:
                margin = self.margin
            else:
                if self.nrows == 1:
                    margin = [50, 10, 50, 30]
                else:
                    margin = [30, 30, 30, 30]
                    row, _ = np.where(grid == idx)
                    if row == 0:
                        margin[0] += 10
                        margin[2] -= 10
                    if row == self.nrows - 1:
                        margin[2] += 10
                        margin[0] -= 10

                # ensure room for the scale bar
                if self.scale_bar:
                    if self.layout in "du":
                        margin[3] += 20
                    elif self.layout in "lr":
                        margin[2] += 20
                margin = tuple(margin)

            # create Cartesian and append to list
            axes = self.canvas.cartesian(
                grid=(self.nrows, self.ncols, idx),
                padding=self.padding,
                margin=margin,
            )
            axes.margin = margin
            self.axes.append(axes)

    def get_canvas(self) -> None:
        """set the canvas with user supplied height width else estimated"""
        # wider than tall
        if self.layout in ("d", "u"):
            minx = 225
            miny = 250
            self.width = self.width if self.width else min(750, minx * self.ncols)
            self.height = self.height if self.height else min(750, miny * self.nrows)
        else:
            minx = 250
            miny = 225
            self.height = self.height if self.height else min(750, minx * self.nrows)
            self.width = self.width if self.width else min(750, miny * self.ncols)
        self.canvas = Canvas(
            height=self.height,
            width=self.width,
        )


# def get_canvas_and_axes(...) -> ...:
#     ...


# class GridSetup:
#     """Return Canvas and Cartesian axes objects to fit a grid of trees."""

#     def __init__(
#         self,
#         nrows,
#         ncols,
#         width,
#         height,
#         layout,
#         margin,
#         padding,
#         scale_bar,
#     ):

#         # style args can include height/width, nrows, ncols, shared,...
#         self.nrows = nrows
#         self.ncols = ncols
#         self.width = width
#         self.height = height
#         self.layout = layout
#         self.margin = margin
#         self.padding = padding
#         self.scale_bar = scale_bar

#         # get .canvas and .axes
#         self.get_tree_dims()
#         self.canvas = toyplot.Canvas(
#             height=self.height,
#             width=self.width,
#         )
#         self.axes = []
#         self.get_axes()
#         # self.get_axes_list()

#     def get_axes(self):
#         """
#         Get a list of axes in the grid shape, and set margins to
#         to make space for optional scale_bars and axes labels.
#         """
#         nplots = self.nrows * self.ncols
#         grid = np.arange(nplots).reshape((self.nrows, self.ncols))

#         for idx in range(nplots):
#             if self.margin:
#                 margin = self.margin
#             else:
#                 if self.nrows == 1:
#                     margin = [50, 10, 50, 30]
#                     # else:
#                     # margin = [50, 20, 50, 20]
#                 else:
#                     margin = [30, 30, 30, 30]
#                     row, _ = np.where(grid == idx)
#                     if row == 0:
#                         margin[0] += 10
#                         margin[2] -= 10
#                     if row == self.nrows - 1:
#                         margin[2] += 10
#                         margin[0] -= 10
#                 # ...
#                 if self.scale_bar:
#                     if self.layout in "du":
#                         margin[3] += 20
#                     elif self.layout in "lr":
#                         margin[2] += 20
#                 margin = tuple(margin)

#             axes = self.canvas.cartesian(
#                 grid=(self.nrows, self.ncols, idx),
#                 padding=self.padding,
#                 margin=margin,
#             )
#             axes.margin = margin
#             self.axes.append(axes)

#     def get_tree_dims(self):
#         """Get height and width if not set by user
#         """
#         # wider than tall
#         if self.layout in ("d", "u"):
#             minx = 225
#             miny = 250
#             self.width = self.width if self.width else min(750, minx * self.ncols)
#             self.height = self.height if self.height else min(750, miny * self.nrows)

#         else:
#             minx = 250
#             miny = 225
#             self.height = self.height if self.height else min(750, minx * self.nrows)
#             self.width = self.width if self.width else min(750, miny * self.ncols)


if __name__ == "__main__":

    # get_multitree_grid()

    grid = Grid(2, 1, None, None, 'r', None, 10, False)
    print(grid)



#!/usr/bin/env python

"""Builtin tree styles.

"""

from typing import Tuple, TypeVar
from dataclasses import dataclass
from toytree.style import TreeStyle

Color = TypeVar("Color")


@dataclass(repr=False, eq=False)
class TreeStyleN(TreeStyle):
    tree_style: str = "n"


@dataclass(repr=False, eq=False)
class TreeStyleS(TreeStyle):
    tree_style: str = "s"
    node_labels: bool = True
    node_mask: bool = False
    node_sizes: int = 16
    tip_labels: bool = True
    use_edge_lengths: bool = False

    def __post_init__(self):
        self.node_style.fill = (0.8274509803921568, 0.8274509803921568, 0.8274509803921568, 1.0)
        self.node_style.stroke = (0.145, 0.145, 0.145, 1.0)
        self.node_style.stroke_width = 1.5


@dataclass(repr=False, eq=False)
class TreeStyleP(TreeStyle):
    """Phylogeny tree_style ..."""
    tree_style: str = "p"
    layout: str = "d"
    edge_type: str = "c"
    edge_widths: Tuple[str] = ("Ne", 2, 6, 1.5)  # special
    node_labels: bool = True
    node_sizes: int = 15
    node_hover: int = False
    node_mask: int = False
    # tip_labels: bool = True
    tip_labels_align: bool = False
    scale_bar: bool = True

    def __post_init__(self):
        self.node_style.stroke = (0.145, 0.145, 0.145, 1.0)
        self.node_style.stroke_width = 1.0
        self.node_labels_style.font_size = 9


@dataclass(repr=False, eq=False)
class TreeStyleO(TreeStyle):
    """Umlaut style."""
    tree_style: str = "o"
    edge_type: str = "c"
    node_sizes: int = 8
    tip_labels_align: bool = True
    # tip_labels: bool = True

    def __post_init__(self):
        self.edge_style.stroke = (0.145, 0.145, 0.145, 1.0)
        self.edge_style.stroke_width = 2
        self.node_style.stroke = (1.0, 1.0, 1.0, 1.0)
        self.node_style.stroke_width = 1.5
        self.node_style.fill = (0.10588235294117647, 0.6196078431372549, 0.4666666666666667, 1.0)


@dataclass(repr=False, eq=False)
class TreeStyleC(TreeStyle):
    """Coalescent style."""
    tree_style: str = "c"
    edge_type: str = "c"
    layout: str = "d"
    scale_bar: bool = True
    node_sizes: int = 7
    node_mask: bool = False

    def __post_init__(self):
        self.tip_labels_style.anchor_shift = 12
        self.edge_style.stroke = (0.145, 0.145, 0.145, 1.0)
        self.edge_style.stroke_width = 2
        self.node_style.stroke = (0.145, 0.145, 0.145, 1.0)
        self.node_style.stroke_width = 1.5
        self.node_style.fill = (0.4, 0.7607843137254902, 0.6470588235294118, 1.0)


@dataclass(repr=False, eq=False)
class TreeStyleR(TreeStyle):
    """R style."""
    tree_style: str = "r"
    node_mask: Tuple[int, int, int] = (0, 1, 1)
    node_colors: Color = "lightgrey"
    node_markers: str = "r2x1"
    node_sizes: int = 16
    node_labels: str = "name"

    def __post_init__(self):
        self.tip_labels_style.font_size: int = 14


@dataclass(repr=False, eq=False)
class TreeStyleB(TreeStyle):
    """Bootstrap support style."""
    tree_style: str = "b"
    edge_widths: Tuple[str] = ("support", 1.5, 4, 1.5)
    # def __post_init__(self):
    #     pass


@dataclass(repr=False, eq=False)
class TreeStyleD(TreeStyle):
    """Dark style."""
    tree_style: str = "d"

    def __post_init__(self):
        self.tip_labels_style.anchor_shift = 8
        self.tip_labels_style.fill = "rgba(90.6%,54.1%,76.5%,1.000)"
        self.edge_style.stroke = "rgba(40.0%,76.1%,64.7%,1.000)"
        self.edge_style.stroke_width = 2
        self.node_labels_style.fill = "rgba(98.8%,55.3%,38.4%,1.000)"
        self.node_style.stroke = None
        self.node_style.stroke_width = 1.5
        self.node_style.fill = "rgba(40.0%,76.1%,64.7%,1.000)"


@dataclass(repr=False, eq=False)
class TreeStyleU(TreeStyle):
    """Unrooted style."""
    tree_style = "u"
    layout = "unrooted"
    use_edge_lengths = False
    node_colors = "white"
    node_sizes = 6
    node_labels = "idx"

    def __post_init__(self):
        self.node_style.stroke = None
        self.node_style.stroke_width = 1.5
        self.node_style.fill = "white"  # do not set defaults that override


def get_base_tree_style_by_name(tree_style: str = "n") -> TreeStyle:
    """Return a base TreeStyle indexed by unique str name prefix"""
    tree_style = tree_style.lower()[0]
    style = STYLE_DICTS[tree_style]
    return (style)()


STYLE_DICTS = {
    "n": TreeStyleN,
    "s": TreeStyleS,
    "p": TreeStyleP,
    "o": TreeStyleO,
    "c": TreeStyleC,
    "d": TreeStyleD,
    "b": TreeStyleB,
    "u": TreeStyleU,
    "r": TreeStyleR,
}


if __name__ == "__main__":


    n = get_base_tree_style_by_name('n')
    c = get_base_tree_style_by_name('d')
    print(n.tip_labels_style.fill)
    print(c.tip_labels_style.fill)


#!/usr/bin/env python

"""Canvas setup functions for single or grids of trees."""

from typing import Optional, Tuple, TypeVar, Union
from loguru import logger

from toyplot.canvas import Canvas
from toyplot.coordinates import Cartesian
from toyplot.mark import Mark

logger = logger.bind(name="toytree")
ToyTree = TypeVar("ToyTree")


def get_linear_width_and_height(mark: Mark) -> Tuple[int, int]:
    """Get width,height to fit the ToyTree Mark given its extents.
    """
    # get space needed for tip labels to not overlap
    ext = mark.extents(['x', 'y'])[1]
    if mark.layout in "rl":
        exts = ext[3] - ext[2]
    else:
        exts = ext[1] - ext[0]
    name_vspace = sum(exts[:mark.ttable.shape[0]]) * 1.15

    # get space needed for tip labels to fit on canvas
    ext = mark.extents(['x', 'y'])[1]
    if mark.layout in "rl":
        exts = ext[1] - ext[0]
    else:
        exts = ext[3] - ext[2]
    name_hspace = max(exts[:mark.ttable.shape[0]])# 2

    # get depth of tree to show edge, nodes, etc.
    tree_depth = max(100, 5 * mark.ttable.shape[0])
    tree_depth = min(300, 5 * mark.ttable.shape[0])

    # add 100 to each dimension for margin; limit w,h = (800, 1000)
    width = tree_depth + name_hspace + 100
    height = name_vspace + 100

    # ... min: (350, 300)
    if mark.layout in "ud":
        width, height = height, width
        height = max(300, min(600, height))
        width = max(300, min(800, width))

    # ... min: (350, 1000)
    else:
        width = max(300, min(800, width))
        height = min(1000, max(275, height))
    return width, height


def get_circular_width_and_height(mark: Mark) -> Tuple[int, int]:
    """TODO."""
    return 500, 500


def get_canvas_and_axes(
    axes: Optional[Cartesian],
    mark: Mark,
    width: Optional[int],
    height: Optional[int],
    padding: int = 15,
    margin: int = 50,
) -> Union[Tuple[Canvas, Cartesian], Tuple[None, Cartesian]]:
    """Get Canvas, Cartesian for a ToyTree drawing."""
    # Create new Carteian to plot tree onto
    if axes is None:
        if mark.layout in "rlud":
            _width, _height = get_linear_width_and_height(mark)
            if width is None:
                width = _width
            if height is None:
                height = _height
        else:
            _width, _height = get_circular_width_and_height(mark)
            if width is None:
                width = _width
            if height is None:
                height = _height

        # create canvas and axes
        canvas = Canvas(height=height, width=width)
        axes = canvas.cartesian(padding=padding, margin=margin)

    # tree is being plotted on an existing set of axes
    else:
        canvas = None
    return canvas, axes


# class CanvasSetup:
#     """Return Canvas and Cartesian axes objects for drawing size.

#     Sets values to style.height and style.width if not present. If a
#     Canvas already exists a set of axes can be entered, instead of
#     being generated anew.
#     """

#     def __init__(self, tree, axes, style):

#         # args includes axes
#         self.tree = tree
#         self.axes = axes
#         self.style = style
#         self.canvas = None
#         self.external_axis = False

#         # get the longest name for dimension fitting
#         self.lname = 0
#         if self.style.tip_labels is not None:
#             self.lname = max([len(str(i)) for i in self.style.tip_labels])

#         # ntips and shape to fit with provided args
#         self.get_canvas_height_and_width()

#         # fills canvas and axes
#         self.get_canvas_and_axes()

#         # ticks for tree and scale_bar
#         if self.style.scale_bar is False:
#             if not self.external_axis:
#                 self.axes.x.show = False
#                 self.axes.y.show = False
#         else:
#             if style.use_edge_lengths:
#                 theight = self.tree.treenode.height
#             else:
#                 # get number of nodes from farthest leaf to root
#                 ndists = self.tree.distance.get_node_distance_matrix(True)
#                 theight = ndists[self.tree.treenode.idx].max()
#             set_axes_ticks_style(theight, self.axes, self.style, True)

#     def get_canvas_height_and_width(self):
#         """Calculate default canvas height&width given N tips and style."""
#         if self.style.layout[0] == "c":
#             radius = max([0] + [i for i in [self.style.height, self.style.width] if i])
#             if not radius:
#                 radius = 400
#             self.style.width = self.style.height = radius
#             return

#         # fit height and width by tree size.
#         if self.style.layout in ("r", "l"):
#             if not self.style.height:
#                 self.style.height = max(275, min(1000, 18 * self.tree.ntips))
#             if not self.style.width:
#                 self.style.width = max(250, min(500, 250 + 5 * self.lname))
#         else:
#             if not self.style.height:
#                 self.style.height = max(250, min(500, 250 + 5 * self.lname))
#             if not self.style.width:
#                 self.style.width = max(350, min(1000, 18 * self.tree.ntips))

#     def get_canvas_and_axes(self):
#         """Sets canvas and axes with dimensions and padding."""
#         if self.axes is not None:
#             self.canvas = None
#             self.external_axis = True
#         else:
#             self.canvas = toyplot.Canvas(
#                 height=self.style.height,
#                 width=self.style.width,
#             )
#             self.axes = self.canvas.cartesian(padding=self.style.padding)


# def set_axes_ticks_style(
#     tree_height: float,
#     axes: Cartesian,
#     style: TreeStyle,
#     only_inside: bool = True,
# ) -> Cartesian:
#     """Return a toyplot Cartesian object with custom tick marks.

#     This gets tick locations first using toyplot.locator.Extended and
#     then sets labels on them using toyplot.locator.Explicit, because
#     we need time scale bar to be non-negative when axes are rotated
#     for trees facing different directions.

#     Note
#     -----
#     Some work is done internally to try to nicely handle floating point
#     precision.

#     Parameters
#     ----------
#     ...
#     style: TreeStyle
#         A TreeStyle object with options for styling axes.
#     only_inside: bool
#         Option used by toyplot.locator.Extended to automatically find
#         tick marks given the data range.
#     """
#     # the axes is either new or passed as an arg, and the scale_bar
#     # arg is True or a (float, int), so we need to style the ticks.
#     if style.layout in ("r", "l"):
#         nticks = max((4, np.floor(style.width / 75).astype(int)))
#         axes.y.show = False
#         axes.x.show = True
#         axes.x.ticks.show = True
#     elif style.layout in ("u", "d"):
#         nticks = max((4, np.floor(style.height / 75).astype(int)))
#         axes.x.show = False
#         axes.y.show = True
#         axes.y.ticks.show = True
#     # e.g., unrooted layout with axes shown (e.g., ts='p')
#     else:
#         # nticks = max((4, np.floor(style.height / 75).astype(int)))
#         nticks = 5
#         axes.x.show = False
#         axes.y.show = False

#     # get tick locator
#     lct = toyplot.locator.Extended(count=nticks, only_inside=only_inside)

#     # get root tree height
#     if style.layout in ("r", "u"):
#         locs = lct.ticks(-tree_height, -0)[0]
#     else:
#         locs = lct.ticks(0, tree_height)[0]

#     # apply unit scaling
#     if style.scale_bar is False:
#         labels = abs(locs.copy())
#     elif isinstance(style.scale_bar, (int, float)):
#         labels = abs(locs / style.scale_bar)
#     else:
#         labels = abs(locs.copy())
#     labels = [np.format_float_positional(i, precision=6, trim="-") for i in labels]

#     # set the ticks locator
#     if style.layout in ("r", "l"):
#         axes.x.ticks.locator = toyplot.locator.Explicit(
#             locations=locs + style.xbaseline,
#             labels=labels,
#         )
#     elif style.layout in ("u", "d"):
#         axes.y.ticks.locator = toyplot.locator.Explicit(
#             locations=locs + style.ybaseline,
#             labels=labels,
#         )
#     # print(locs, labels)
#     return axes


if __name__ == "__main__":

    import toytree
    t0 = toytree.tree("https://eaton-lab.org/data/Cyathophora.tre")
    t0 = t0.root("~.*prz*")
    c0, _, _ = t0.draw()
    print(c0.width, c0.height)

    t1 = toytree.rtree.unittree(26)
    c1, _, _ = t1.draw()
    print(c1.width, c1.height)

    toytree.utils.show([c0, c1])




#!/usr/bin/env python

"""Example datasets for testing and demonstration.


toytree.data.newick
toytree.dtaa.newick_multitree
toytree.data.nexus
toytree.data.nexus_multitree
toytree.data.nhx
toytree.data.nhx_mb
toytree.data.nhx_beast
toytree.data.distance_matrix
toytree.data.sequence_alignment
"""

from toytree.data._src.get_node_data import get_node_data
from toytree.data._src.set_node_data import set_node_data
from toytree.data._src.expand_node_mapping import expand_node_mapping
# from toytree.data._src.transform import normalize_values



#!/usr/bin/env python

"""A horizontal scrolling toyplot.Canvas.

This is currently used by the ipcoal and shadie packages to
visualize tree sequences simulated in SLiM and/or msprime using
the `ToyTreeSequenceDrawing` class.
"""

import toyplot

class ScrollableCanvas(toyplot.Canvas):
    """Canvas subclass with horizontal scrolling on large widths.

    This is returned as part of a ToyTreeSequenceDrawing in the typical
    three-part tuple of a toyplot drawing as (Canvas, axes, mark).
    """
    def _repr_html_(self):
        return toyplot.html.tostring(
            self, style={"text-align": "center", "width": f"{self.width}px"}
        )



#!/usr/bin/env python

"""Rewrite of the container drawing class.

Note
----
Ne values are fixed within intervals. The steps to drawing are:

1. Get horizontal spacing of species tree tips based on Ne values.

2. Begin drawing tip lineages in post-order traversal. This start by
creating an Interval for each tip in a cherry, and for their ancestor.
The end positions of the tips is designed to lean towards their parent,
whose positionCreate a new Interval for the cherry parent.

Ideas
------
container = mark (square, angled, styled?)
embedding = mark (must fit within container?, allow migration?)
"""

# pylint: disable=invalid-name

from __future__ import annotations
from typing import List, Dict, Optional, Mapping, Sequence, Union, Tuple
from functools import cached_property
from dataclasses import dataclass, field
from loguru import logger
import numpy as np
from pandas import DataFrame

import toyplot
from toytree.style import get_range_mapped_feature, get_color_mapped_values
from toytree.utils.src.embedding import get_genealogy_embedding_table
from toytree.core import ToyTree, Canvas, Cartesian, Mark, Node
from toytree.color import ColorType

logger = logger.bind(name="toytree")
# ToyTree = TypeVar("ToyTree")


@dataclass
class Interval:
    r"""A single interval of the species tree model.

         pxb0\---------------------/pxb1
            xt0________xt1  ...   ...              y1
             /       /       \     \               |
            /       /         \     \              |
        xb0/_______/xb1        ...   ...           y0
               |
              xm0

    Parameters
    ----------
    xpos: float
        The x position of the lower bound of this interval. This is the
        original left position, but may be updated and extended further
        left if blend=True.
    ypos: float
        The y position (height) of the lower bound of this interval.
    up: Interval or None
        An Interval class that is parent to this one.
    children: List
        A list of Intervals directly descended from this one.
    left: bool
        If True this lineage is the left descendant of its parent.
    blend: bool
        If True then interval bottoms are fit to the width of their
        children intervals to make the tree look less blocky.
    """
    width: float
    height: float
    xpos: float = 0
    ypos: float = 0
    neff: float = 0
    left: bool = False
    blend: bool = False
    up: Optional[Interval] = field(default=None, init=False, repr=False)
    children: List[Interval] = field(default_factory=list, init=False, repr=False)
    data: Dict[str, float] = field(default_factory=dict, init=False, repr=False)

    @cached_property
    def xb0(self):
        """Return the left x-position on the lower bound"""
        if self.children and self.blend:
            return min([i.xt0 for i in self.children])
        return self.xpos

    @cached_property
    def xmid(self):
        """Return the midpoint x-position on lower bound"""
        return self.xpos + (self.width / 2.)

    @cached_property
    def xb1(self):
        """Return the right x-position on the lower bound"""
        if self.children and self.blend:
            return max([i.xt1 for i in self.children])
        return self.xb0 + self.width

    @cached_property
    def xt0(self):
        """Return the left x-position on the upper bound"""
        if not self.up:
            return self.xb0
        if not self.left:
            return self.up.xmid - self.width
        return self.up.xmid

    @cached_property
    def xt1(self):
        """Return the right x-position on the upper bound"""
        if not self.up:
            return self.xb1
        if not self.left:
            return self.up.xmid
        return self.up.xmid + self.width

    @cached_property
    def y0(self):
        """Return the y-position of the lower bound"""
        return self.ypos

    @cached_property
    def y1(self):
        """Return the y-position of the upper bound"""
        return self.ypos + self.height

    # @cached_property
    # def xbounds(self):
    #     """Return the bounding coordinates (xb0, xb1, xt0, xt1)"""
    #     return (self.xb0, self.xb1, self.xt0, self.xt1)

    # @cached_property
    # def ybounds(self):
    #     """Return the bounding coordinates (y0, y0, y1, y1)"""
    #     return (self.y0, self.y0, self.y1, self.y1)

    # @cached_property
    def slope(self, position: float = None):
        """Return the slope (delta x / delta y) for use in :func:is_inside()"""
        if not self.up:
            return 0.
        if not position:
            position = (self.xt1 - self.xt0) / 2.
        # dist to right of xb0 / length of xb
        prop_bottom = (position - self.xb0) / (self.xb1 - self.xb0)
        length_on_top = prop_bottom * (self.xt1 - self.xt0)
        pos_at_same_prop_on_top = self.xt0 + length_on_top
        return (pos_at_same_prop_on_top - position) / self.height

    def is_inside(
        self,
        xpos: float,
        ypos: float,
        padding: float = 0.0,
    ):
        """Return True if coordinate position is inside the interval.

        This is used to avoid overlaps among intervals, and to ensure
        embedded gene trees do not extend outside of of the interval.
        The padding parameter can be used to further limit embeddings
        from coming too close to the edge, or similarly, margin will
        pad the outside to limit how close intervals are to each other.
        """
        ypass = self.y0 <= ypos <= self.y1
        xmin_at_ypos = self.xb0 + padding + (ypos - self.y0) * self.slope()
        xmax_at_ypos = self.xb1 - padding + (ypos - self.y0) * self.slope()
        # logger.info(f"xmin_at_ypos: {xmin_at_ypos}")
        # logger.info(f"xmax_at_ypos: {xmax_at_ypos}")
        xpass = xmin_at_ypos <= xpos <= xmax_at_ypos
        return xpass & ypass

    def is_overlapping(
        self,
        xpos: float,
        ypos: float,
        border: float = 0.0,
    ):
        """Return True if coordinate position is inside the interval.

        This is used to avoid overlaps among intervals, and to ensure
        embedded gene trees do not extend outside of of the interval.
        The padding parameter can be used to further limit embeddings
        from coming too close to the edge, or similarly, margin will
        pad the outside to limit how close intervals are to each other.
        """
        margin: float = 0.0


@dataclass
class Container:
    """Container to visualize a genealogy embedded in a species tree.

    A Container drawing is composed of multiple Interval class
    instances that each have coordinates, and reference each other
    in a tree structure (.up, .children), and can be accessed from
    the Container's .intervals dict.

    Parameters
    ----------
    species_tree: toytree.ToyTree
        A ToyTree that optionally has 'Ne' data set to nodes to use for
        relative interval widths. If no Nes then set to equal widths.
    gene_tree: toytree.ToyTree
        A genealogy to embed in the species tree container.
    spacing: float
        A value to increase spacing between intervals to avoid overlap
        among angled intervals of the tree.
    min_width: float
        A value for the minimum width of intervals after Ne values are
        re-scaled and binned make visualizations fit more simply on
        the canvas.
    max_width: float
        See min_width.
    """
    tree: Union[ToyTree, int]
    spacing: float = 1.5
    blend: bool = False

    # Attributes
    intervals: Dict[int, Interval] = field(default_factory=dict, init=False)
    """: Dict of Interval class objects used for coordinates."""
    min_width: int = 2
    """: Minimum width of an interval, used to help fit drawing on canvas."""
    max_width: int = 8
    """: Maximum width of an interval, used to help fit drawing on canvas."""
    canvas: Canvas = None
    axes: Cartesian = None

    def __post_init__(self):
        """Assign coordinates to Intervals in the species tree model.

        The width of intervals and spacing can be tuned by the user.
        """
        # single population coalescent
        if isinstance(self.tree, (int, float)):
            self.intervals[0] = Interval(
                width=8,
                height=np.inf,
                xpos=0,
                ypos=0,
                neff=self.tree,
                blend=self.blend
            )
            tree = ToyTree(Node(name="0"))
            tree[0].Ne = self.tree
            self.tree = tree
            return

        # species tree model
        # get the widths of the tip intervals
        if "Ne" not in self.tree.features:
            self.tree = self.tree.set_node_data("Ne", default=1000)
        widths = get_range_mapped_feature(
            self.tree,
            "Ne",
            min_value=self.min_width,
            max_value=self.max_width,
            nan_value=self.min_width,
        )

        # traverse in idx order (tips then post-order for internal)
        for idx, node in enumerate(self.tree):

            # tips get x-position from widths starting from zero, whereas
            # internal nodes get their midpoint from child nodes and then
            # expand to sides for width.
            if not node.is_leaf():
                xmids = [self.intervals[child.idx].xmid for child in node.children]
                xmid = sum(xmids) / len(xmids)
                xpos = xmid - (widths[idx] / 2.)
            else:
                xpos = 0 if not idx else self.intervals[idx - 1].xb1 + self.spacing

            # store the interval
            self.intervals[idx] = Interval(
                width=widths[idx],
                height=node.dist,
                xpos=xpos,
                ypos=node.height,
                neff=node.Ne,
                blend=self.blend
            )

            # set child-parent relationships
            self.intervals[idx].children = []
            for child in node.children:
                cidx = child.idx
                self.intervals[cidx].up = self.intervals[idx]
                if self.intervals[idx].children:
                    self.intervals[cidx].left = True
                self.intervals[idx].children.append(self.intervals[cidx])

    def setup_drawing(self, **kwargs) -> Tuple[Canvas, Cartesian]:
        self.canvas = toyplot.Canvas(
            width=kwargs.pop('width', 400),
            height=kwargs.pop('height', 400),
        )
        self.axes = self.canvas.cartesian()
        self.axes.x.show = False
        self.axes.y.ticks.show = True
        return self.canvas, self.axes

    def draw_intervals(
        self,
        container_fill: ColorType,
        container_fill_opacity: float,
        container_fill_opacity_alternate: bool = False,
    ):
        """Return a toyplot drawing of the container.

        """
        # do nothing if single population
        if self.tree.ntips == 1:
            return

        # setup canvas and axes
        # add fill intervals
        marks = []
        for node in self.tree.traverse("levelorder"):
            idx = node.idx
            ival = self.intervals[idx]

            # hover-over information for interval
            title = (
                f"idx={idx}\n"
                f"Ne={ival.neff:.2g}\n"
                f"Tc={ival.height / (2 * ival.neff):.3g}\n"
                f"Tg={ival.height:.2g}"
            )

            # optional alternating stripes of color opacity
            ival.fill = container_fill
            ival.fill_opacity = container_fill_opacity
            if container_fill_opacity_alternate:
                if not ival.up:
                    ival.stripe = 1
                    ival.opacity = container_fill_opacity
                elif ival.up.stripe == 1:
                    ival.stripe = 0
                    ival.opacity = container_fill_opacity / 2
                else:
                    ival.opacity = container_fill_opacity
                    ival.stripe = 1

            # do not draw root yet
            if node.is_root():
                continue

            # draw other nodes (angled left or right)
            if ival.left:
                mark = self.axes.fill(
                    [ival.xb0, ival.xt0, ival.xb1, ival.xt1],
                    [ival.y0, ival.y0, ival.y0, ival.y1],
                    [ival.y0, ival.y1, ival.y1, ival.y1],
                    title=title,
                    color=ival.fill,
                    opacity=ival.opacity,
                )
            else:
                mark = self.axes.fill(
                    [ival.xt0, ival.xb0, ival.xt1, ival.xb1],
                    [ival.y1, ival.y0, ival.y0, ival.y0],
                    [ival.y1, ival.y1, ival.y1, ival.y0],
                    title=title,
                    color=ival.fill,
                    opacity=ival.opacity,
                )
            marks.append(mark)

        # style the drawing
        return self.canvas, self.axes, marks

    def draw_root_interval(
        self,
        container_fill: ColorType,
        container_fill_opacity: float,
        height: float = 100,
    ):
        """
        # draw the root (no angle) as 20% of treeheight by default.
        # To draw this height to fit the gene tree requires knowing
        # the gene tree's root height...
        """
        ival = self.intervals[max(self.intervals)]
        title = (
            f"idx={max(self.intervals)}\n"
            f"Ne={ival.neff:.2g}\n"
            f"Tc={ival.height / (2 * ival.neff):.3g}\n"
            f"Tg={ival.height:.2g}"
        )
        mark = self.axes.fill(
            [ival.xb0, ival.xb0, ival.xb1, ival.xb1],
            [ival.y0] * 4,
            [ival.y0, ival.y0 + height, ival.y0 + height, ival.y0],
            title=title,
            color=container_fill,
            opacity=container_fill_opacity,
        )
        return mark
        # tee = len(str(int(self.maxheight)))
        # tze = (1 * 10 ** (tee - 2))
        # tma = self.maxheight
        # trd = round(tma / tze) * tze
        # print(tee, tze, tma, trd)
        # axes.y.ticks.locator = toyplot.locator.Explicit(
        #     np.linspace(0, trd, 6).astype(int)[:-1],
        #     ["{:.0f}".format(i) for i in np.linspace(0, trd, 6).astype(int)][:-1],
        # )
        # self.axes.y.domain.max = self.maxheight + (self.maxheight * 0.01)
        # self.axes.y.domain.min = -self.maxheight * 0.05

    def draw_stroke(
        self,
        container_stroke,
        container_stroke_width,
        container_stroke_opacity,
    ):
        # TODO
        pass


@dataclass
class EmbeddingPlot:
    """Embed a gene tree in a container.

    This ...
    """
    species_tree: Union[ToyTree, int]
    gene_tree: ToyTree
    imap: Mapping[str, Sequence[str]] = field(default_factory=dict)
    blend: bool = False
    spacing: float = 1.5
    min_width: float = 2.
    max_width: float = 8.
    table: DataFrame = field(default=None, repr=False, init=False)
    container: Container = field(default=None, repr=False, init=False)

    # drawing components
    canvas: Canvas = field(default=None, repr=None, init=False)
    axes: Cartesian = field(default=None, repr=None, init=False)
    marks: List[Mark] = field(default_factory=list, repr=False, init=False)
    _xpoints: List[float] = field(default_factory=list, repr=None, init=False)
    _ypoints: List[float] = field(default_factory=list, repr=None, init=False)
    _labels: List[str] = field(default_factory=list, repr=None, init=False)

    def __post_init__(self):
        # create a container of connected intervals and a ToyTree w/ Ne values
        self.container = Container(
            self.species_tree, blend=self.blend, spacing=self.spacing,
            min_width=self.min_width, max_width=self.max_width,
        )
        self.species_tree = self.container.tree

        # ensure imap
        self.imap = self.imap if self.imap else {'0': self.gene_tree.get_tip_labels()}

        # get a genealogy embedding table
        self.table = get_genealogy_embedding_table(
            self.species_tree, self.gene_tree, self.imap)

    def get_node_coordinates(self, container_root_height: int) -> None:
        """Sample position of each Node at interval (start, end/coal)

        """
        for sidx, ival in self.container.intervals.items():
            tab = self.table[self.table.st_node == sidx]

            if sidx < self.species_tree.ntips:
                # store evenly spaced intervals of tip nodes
                nedges = tab.iloc[0].nedges
                xpos = np.linspace(ival.xb0, ival.xb1, nedges + 2)[1:-1]
                for nidx, pos in zip(tab.iloc[0].edges, xpos):
                    ival.data[nidx] = [pos]
                self._xpoints.append(xpos)
                self._ypoints.append(np.repeat(ival.y0, nedges))
                self._labels.append([f"node={nidx}" for nidx in tab.iloc[0].edges])

            else:
                # get positions at interval intersections from left child
                cidx = self.species_tree[sidx].children[0].idx
                tabl = self.table[self.table.st_node == cidx]
                nedges = tabl.iloc[-1].nedges
                edges = tabl.iloc[-1].edges
                # print(sidx, 'l', edges, ival.children[0].data)
                edges = sorted(edges, key=lambda x: ival.children[0].data[x][0])
                xposl = np.linspace(ival.xb0, ival.xmid, nedges + 2)[1:-1]
                for nidx, pos in zip(edges, xposl):
                    ival.data[nidx] = [pos]

                # get positions at interval intersections from right child
                cidx = self.species_tree[sidx].children[1].idx
                tabr = self.table[self.table.st_node == cidx]
                nedges = tabr.iloc[-1].nedges
                edges = tabr.iloc[-1].edges
                edges = sorted(edges, key=lambda x: ival.children[1].data[x][0])
                # print(sidx, 'r', edges, ival.children[1].data)
                xposr = np.linspace(ival.xmid, ival.xb1, nedges + 2)[1:-1]
                for nidx, pos in zip(edges, xposr):
                    ival.data[nidx] = [pos]
                # self._xpoints.append(np.concatenate([xposl, xposr]))
                # self._ypoints.append(np.repeat(ival.y0, xpos.size))

            # store midpoint positions for coalescence nodes
            for idx in tab.index[:-1]:

                # get who coalesced into whom
                cidxs = set(tab.loc[idx].edges) - set(tab.loc[idx + 1].edges)
                nidx = set(tab.loc[idx + 1].edges) - set(tab.loc[idx].edges)

                # get midpoint position on slope
                shifted = []
                for c in cidxs:
                    pos = ival.data[c][0]
                    slope = ival.slope(pos)
                    # length of edge in this interval
                    node = self.gene_tree[c]
                    dist = node.height + node.dist - ival.ypos
                    shifted.append(pos + (dist * slope))
                    # shifted.append(ival.data[c][0] + shift)
                xpos = sum(shifted) / 2.

                # store the start position of the ancestor coal node
                nidx = nidx.pop()
                ival.data[nidx] = [xpos]

                # store the end positions of nodes that coalesced
                for cidx in cidxs:
                    ival.data[cidx].append(xpos)

                # only store point for plotting if inside plotted intervals
                ypos = tab.loc[idx].stop
                if ypos <= container_root_height:
                    self._xpoints.append([xpos])
                    self._ypoints.append([ypos])
                    self._labels.append([f"node={nidx}"])

    def add_connecting_lines(
        self,
        edge_stroke,
        edge_stroke_width,
        edge_stroke_inherit: bool,
        edge_stroke_opacity: float,
        edge_samples: int,
        edge_variance: int,
        container_root_height: int
    ):
        """...
        """
        # scaling the wiggle. Allow 50 wiggles from tip to root, sample
        # number of wiggles per edge according to the proportion of len.
        if (edge_samples < 2) | (edge_variance <= 0):
            edge_samples = 0
            edge_variance = 1

        # colormap
        # edge_stroke,
        cmap = get_color_mapped_values(range(self.species_tree.ntips), cmap="Dark2")

        # draw edges within each interval
        for sidx, ival in self.container.intervals.items():
            # stroke = stroke if stroke is not None else COLORS2[sidx]
            table = self.table[self.table.st_node == sidx]
            edges_in_interval = set.union(*table.edges.apply(set).tolist())
            for e in edges_in_interval:
                if e != self.gene_tree.treenode.idx:
                    prop = self.gene_tree[e].dist / self.gene_tree.treenode.height
                    nsamps = int(prop * edge_samples) + 2

                    # get x coords
                    x0 = ival.data[e][0]
                    x1 = ival.data[e][1] if len(ival.data[e]) == 2 else ival.up.data[e][0]

                    # get y coords
                    y0 = self.gene_tree[e].height
                    if y0 < table.iloc[0, 0]:
                        y0 = table.iloc[0, 0]
                    if y0 >= container_root_height:
                        continue
                    y1 = self.gene_tree[e].up.height
                    if y1 > table.iloc[-1, 1]:
                        y1 = table.iloc[-1, 1]

                    # if sample does not coalesce in final interval
                    if y1 > container_root_height:
                        y1 = container_root_height
                        xdiff = x0 - x1
                        x1 = x0 - (xdiff / 2.)

                    # override color to black if internal not monophyletic
                    kidx = check_monophyly(self.gene_tree[e].get_leaf_names(), self.imap)
                    if kidx == -1:
                        stroke = "black"
                    else:
                        stroke = cmap[kidx]

                    # add wiggle to x coords
                    ypos = np.linspace(y0, y1, nsamps)
                    xpos = np.linspace(x0, x1, nsamps)
                    xpos[1:-1] += np.random.uniform(-edge_variance, edge_variance, size=nsamps - 2)
                    self.axes.plot(
                        xpos, ypos,
                        color=stroke,
                        stroke_width=edge_stroke_width,
                        # stroke_opacity=stroke_opacity
                    )

    def add_tip_labels(self):
        for sidx, ival in self.container.intervals.items():
            if sidx < self.species_tree.ntips:
                xpos = [ival.data[i][0] for i in ival.data if i < self.gene_tree.ntips]
                names = [i for i in ival.data if i < self.gene_tree.ntips]
                names = [self.gene_tree[i].name for i in names]
                self.axes.text(
                    xpos, [0] * len(xpos), names,
                    angle=-90,
                    style={
                        "font-size": 12,
                        "fill": "black",
                        "-toyplot-anchor-shift": 20,
                    },
                )

    def add_nodes(
        self,
        node_size,
        node_fill,
        node_fill_opacity,
        node_stroke,
        node_stroke_width,
    ):
        node_stroke = "none" if node_stroke is None else node_stroke
        self.axes.scatterplot(
            np.concatenate(self._xpoints),
            np.concatenate(self._ypoints),
            title=np.concatenate(self._labels),
            size=node_size,
            mstyle={
                "fill": node_fill,
                "fill-opacity": node_fill_opacity,
                "stroke": node_stroke,
                "stroke-width": node_stroke_width,
            }
        )

    def add_container(
        self,
        container_width,
        container_height,
        container_fill,
        container_fill_opacity,
        container_fill_opacity_alternate,
        container_stroke,
        container_stroke_opacity,
        container_stroke_width,
        container_root_height,
    ):
        self.canvas, self.axes = self.container.setup_drawing(
            width=container_width,
            height=container_height,
        )

        # draw the species tree w/ root interval depending on user option
        if self.species_tree.ntips == 1:
            self.container.draw_root_interval(
                container_fill=container_fill,
                container_fill_opacity=container_fill_opacity,
                height=container_root_height,  # self.gene_tree.treenode.height
            )
        else:
            self.container.draw_intervals(
                container_fill=container_fill,
                container_fill_opacity=container_fill_opacity,
                container_fill_opacity_alternate=container_fill_opacity_alternate,
            )
            self.container.draw_root_interval(
                container_fill=container_fill,
                container_fill_opacity=container_fill_opacity,
                height=container_root_height,
            )
        self.container.draw_stroke(
            container_stroke,
            container_stroke_opacity,
            container_stroke_width,
        )

    def draw(
        self,
        # container_blend_intervals: bool = False, # already applied
        container_width: int = 350,
        container_height: int = 300,
        container_fill: ColorType = "black",
        container_fill_opacity: float = 0.25,
        container_fill_opacity_alternate: bool = True,
        container_stroke: ColorType = "black",
        container_stroke_opacity: float = 1.0,
        container_stroke_width: float = 2,
        container_root_height: Union[int, bool] = True,
        node_fill: ColorType = "black",
        node_fill_opacity: float = 1.0,
        node_stroke: ColorType = None,
        node_stroke_width: int = 1,
        node_size: int = 5,
        edge_stroke: ColorType = "black",
        edge_stroke_width: int = 2,
        edge_stroke_inherit: bool = True,
        edge_stroke_opacity: float = 1.0,
        edge_samples: int = 10,
        edge_variance: float = 0.0,
        tip_labels_size: int = 10,
    ):
        # single population drawings fit the entire interval
        if self.species_tree.ntips == 1:
            if container_root_height is True:
                container_root_height = self.gene_tree.treenode.height
        else:
            if container_root_height is True:
                container_root_height = (
                    self.gene_tree.treenode.height - self.species_tree.treenode.height)

        # draw species tree container
        self.add_container(
            container_width, container_height,
            container_fill, container_fill_opacity, container_fill_opacity_alternate,
            container_stroke, container_stroke_opacity, container_stroke_width,
            container_root_height,
        )

        # sample positions at interval intersections
        self.get_node_coordinates(self.species_tree.treenode.height + container_root_height)

        # draw the figure
        self.add_connecting_lines(
            edge_stroke=edge_stroke,
            edge_stroke_width=edge_stroke_width,
            edge_stroke_inherit=True,
            edge_stroke_opacity=1.0,
            edge_samples=edge_samples,
            edge_variance=edge_variance,
            container_root_height=self.species_tree.treenode.height + container_root_height,
        )
        self.add_tip_labels()
        self.add_nodes(
            node_size,
            node_fill,
            node_fill_opacity,
            node_stroke,
            node_stroke_width,
        )


def check_monophyly(names, imap) -> int:
    for idx, key in enumerate(imap):
        vals = imap[key]
        if all(i in vals for i in names):
            return idx
    return -1


if __name__ == "__main__":

    import ipcoal
    import toytree
    toytree.set_log_level("INFO")

    # tree = toytree.rtree.unittree(ntips=4, treeheight=1e6, seed=123)
    # tree = tree.mod.edges_slider(seed=333)
    # model = ipcoal.Model(tree, Ne=1e5)
    # model.sim_trees(5)
    # gtree = toytree.tree(model.df.genealogy[0])

    S = toytree.rtree.imbtree(3, treeheight=5e3)
    S.set_node_data("Ne", default=3000, inplace=True)
    M = ipcoal.Model(S, nsamples=3, seed_trees=123)
    M.sim_trees(10)
    G = toytree.tree(M.df.genealogy[0])
    IMAP = M.get_imap_dict()

    E = EmbeddingPlot(S, G, IMAP, blend=1)
    E.sample_intersection_positions()
    E.add_coal_points()

    # con = Container(model.tree, blend=True)
    # print(con.intervals)

    # c, a, m = con.draw(width=400, height=250, color='black', stripe=True, opacity=0.33);
    # a.x.show = True
    # toytree.utils.show(c)
